program pascalv14 ;

{$APPTYPE CONSOLE}
{$mode DELPHI}

uses
  SysUtils;

  (* Pascal compiler for M-code interpreter V1.4 *)

  (* Original version : Mark D. Rustad 1977,
                        hacked version of P2-compiler
     Vax-version      : Hans Otten  1984/1985
                        some features P4-compiler restored
     Turbo Pascal/    : Hans otten, 1995 -2020
     Delphi/
     Freepascal
*)

  (* Last update      : 14 aug 2020 Hans Otten      *)

  (* V1.0 Compiler based upon portable P2 compiler
     written by N.Wirth et al Zurich E.T.H.            *)

  (* Additions for version 1.1 :

     - restructured
     - documented
     - upper- and lowercase accepted in identifiers
     - tab-characters accepted and expanded
     - form-feed character accepted and replaced with blank
     - listing file output
     - HALT and EXIT standard procedure added
     - compile time boundary check on sets
       in types and constants
     - beside ELSE also OTHERWISE in CASE added
     - Error numbers renumbered and messages added
     - error in generating ENT instructions with parameter
       list corrected
     - error in generating correct code for set inclusion
       corrected in EXPRESSION and in CONDGEN
     - error in generating code for Read procedure if
       undeclared identifier used corrected

     Additions for version 1.2 :

     - filetype text added in standard names and read and write
     - standard procedures reset and rewrite added including
       file-open non-standard extension
     - standard function status added for file-variables
     - standard procedure CLOSE added for file-variables
     - standard input file keyboard (not buffered, no echo) added
     - error in generating access to 1 byte parameters lod1
     - error in WriteProc checking types with Comptypes corrected
     - error in handling constant string array of char in
       procedure/function call in CallNonStandard corrected

     Additions for version 1.3 :

     - standard functions SUCC and PRED added
     - array of character constants and variables of any length
       may be assigned to variables of type (packed) array of char
       if length is less or more than nr of characters in array
       The result is padded with blanks or truncated
     - character may be assigned to array of char,
       result padded with blanks
     - bug in body (complex structures formal type corrected

     Additions for V1.4

     Turbo/Borland Pascal file open/close hacked in
     Corrected P4 -> P2 origin

       *)

  (* Constants *)

CONST
(***)

  displimit  =    15 ;
  maxlevel   =    12 ;

  (* size of store accessible by code *)
  maxaddr    = 32767 ;

  (* size in byte in core of variables *)
  (* 8 bytes for set, gives 64 possible members *)

  intsize    =     2 ; (* integer   *)
  charsize   =     1 ; (* character *)
  boolsize   =     1 ; (* boolean   *)
  setsize    =     8 ; (* set       *)
  ptrsize    =     2 ; (* pointer   *)


  strglgth   =     8 ; (* strings stored in blocks of 8 char's *)
  alphalen   =    40 ; (* general purpose string length *)

  lcaftermarkstack =     6 ; (* stack frame for interpreter *)

  maxint     = 32767 ; (* 16-bit's two's complement machine *)
  setmax     =    63 ; (* set has maximal 63 members *)
  maxstandrd =    11 ; (* number of standard functions/procedures *)
  maxcode    =    29 ; (* codebuffer maximum 30 bytes *)
  maxpage    =    56 ; (* maximum number of lines on listing file *)
  maxfiles   =     7 ; (* maximum number of files in use total *)
  maxchcnt   =   120 ; (* maximum input line-length *)
  maxerlines =    10 ; (* maximum number of errors in a line *)
  maxermsg   =   200 ; (* maximum number of error messages *)

  (* ASCII character Control Constants *)

  atab     = 9   ; (* tab-character  *)
  aform    = 12  ; (* form-feed *)


TYPE
(**)

  (* basic symbols *)

  symbol  =  (
               ident,
               intconst,
               realconst,
               stringconst,
               notsy,
               mulop,
               addop,
               relop,
               lparent,
               rparent,
               lbrack,
               rbrack,
               comma,
               semicolon,
               period,
               arrow,
               colon,
               becomes,
               constsy,
               typesy,
               varsy,
               funcsy,
               progsy,
               procsy,
               setsy,
               packedsy,
               arraysy,
               recordsy,
               externsy,
               forwardsy,
               beginsy,
               ifsy,
               casesy,
               repeatsy,
               whilesy,
               forsy,
               endsy,
               elsesy,
               untilsy,
               ofsy,
               dosy,
               tosy,
               downtosy,
               thensy,
               othersy

                 ) ;

  operator = (
               mul,
               andop,
               idiv,
               imod,
               plus,
               minus,
               orop,
               ltop,
               leop,
               geop,
               gtop,
               neop,
               eqop,
               inop,
               noop ) ;

   setofsys  =  set of symbol      ;

   intset    =  set of 0 .. setmax ;


   strng = packed array [ 1 .. strglgth ] of 0 .. 127  ;

   (* Constant type descriptor *)

   constptr = ^Constant ;
   Constant = RECORD
                 CASE integer OF

                    (* set *)
                    0 : (pval   : intset ) ;

                    (* String *)
                    1 : ( slgth : 0 .. strglgth ;
                          (* linked list pointer *)
                          sptr  : constptr ;
                          sval  : strng)
                  END ;

   (* value of constant, can be simple integer or complex *)

   valu  =  RECORD
              CASE integer OF
                0 : ( ival : integer ) ;
                1 : ( valp : constptr     )
               END ;


  (* data structures *)

  levrange  = 0 .. maxlevel ;
  addrrange = 0 .. maxaddr  ;

  structform = (
                 scalar,
                 subrange,
                 pointer,
                 power,
                 arrays,
                 records,
                 files
                           ) ;

  declkind  =  (
                 standard,
                 declared
                           ) ;

  (* identifier description *)

  identptr= ^identifier ;

  structptr = ^structure  ;

  structure = PACKED RECORD
                size  : addrrange ;
                stype : structptr ;
                CASE form : structform OF
                  subrange : ( min, max   : valu       ) ;
                  arrays   : ( indextype  : structptr  ) ;
                  records  : ( fstfld     : identptr   )
               END ;

   (* names, identifier-classes  *)

   idclass = ( types,
               konst,
               vars,
               field,
               proc,
               func
                       ) ;

   setofids = set of idclass ;

   idkind = (  actual,
               formal
                       ) ;

   alpha = packed array [ 1 .. strglgth ] of char ;
   alphastring = packed array [ 1 .. alphalen ] of char ;


   identifier = PACKED RECORD
                  name   : alpha ;
                  llink,
                  rlink  : identptr  ;   (* to build binary tree *)
                  idtype : structptr ;
                  next   : identptr  ;   (* to build linked list *)
                  CASE klass : idclass OF
                    konst : ( values : valu )   ;
                    vars  : ( vkind : idkind    ;
                              vlev  : levrange  ;
                              vaddr : addrrange ) ;
                    proc, func :
                      ( CASE pfdeckind : declkind OF
                          standard : ( key : 1 .. maxstandrd ) ;
                          declared : ( pflev   : levrange ;
                                       pfname  : integer  ;
                                       forwdecl ,
                                       externl : boolean  )  )
               END ;

   disprange = 0 .. displimit ;

   (* Expressions *)

   attrkind = (
                 cst,
                 varbl,
                 expr
                       ) ;

   vaccess =  (
                drct,
                indrct
                       ) ;

   (* result of expression evaluation *)

   attr    =  RECORD
                typtr : structptr ;
                CASE kind : attrkind OF
                  cst   : ( cval : valu ) ;
                  varbl : ( access : vaccess  ;
                            level  : levrange ;
                            dplmt  : addrrange  )
               END ;

 VAR
(***)

   (* file related Variables *)

   infile, outfile,listfile : text ;
   filename : alphastring ;

   (* Variables returned by source program scanner InSymbol *)

   sy   : symbol     ;          (* Last symbol found *)
   op   : operator   ;          (* Classification of last symbol *)
   val  : valu       ;          (* Value of last symbol *)
   lgth : integer    ;          (* Length of last string constant *)
   id   : alpha      ;          (* Last identifier,possibly truncated *)
   kk   : 0  .. 8    ;          (* Nr's of chars in last identifier *)
   ch   : char       ;          (* Last character read *)
   ich  : integer    ;          (* ASCII value of last character *)

   (*  Line Variables *)

   linelength,
   chcnt        : 0 .. maxchcnt ;
   linpos       : integer ;
   line         : packed array [ 1 .. maxchcnt ] of char ;
   linecount    : integer ;

   (* listing file counters *)
   listlines, listpages : integer ;

   (* Counters *)

   lc,
   ic   : addrrange   ;     (* Data location and instruction counters *)
   icn,
   sumcheck : integer ;     (* Checksum *)

   codebuf  : array [ 0 .. maxcode ] OF 0 .. 255 ;

   nproc    : integer ;     (* Next procedure number *)
   i,
   mxintio  : integer ;

   (* Switches *)

   declpart,                (* Declaration part *)
   prterr  : boolean ;      (* To allow forward reference in pointer
                               type declaration by suppressing error
                               message    *)
   pfcttest : boolean ;     (* Detect function Calls in
                               parameter lists *)
   errflag  : boolean ;     (* Compiler Error flag *)

   (* Pointers *)

   intptr,
   charptr,
   boolptr,
   nilptr,
   fileptr  : structptr  ;     (* Pointers to standard identifiers *)

   utypptr,
   ucsptr,
   uvarptr,
   ufldptr,
   uprcptr,
   ufctptr  : identptr ;       (* Pointers for undeclared identifiers *)
   fwptr    : identptr ;       (* Head of chain of forw decl type ids *)
   progptr  : identptr ;        (* Pointer to program *)

   (* Bookkeeping of declaration levels *)

   level   : levrange  ;    (* Current static level *)
   disx    : disprange ;    (* Level of last ID searched by Searchid*)
   top     : disprange ;    (* Top of display *)
   savetop : disprange ;    (* Save top of display *)

   (* Files are given a number internal to compiler and interpreter,
      stored in identifier description vaddr-field.
      Input and output predeclared with filenumbers resp 1 and 2
      User-files start with 3 as declared in Initialize           *)

   filenumber :integer ;

   (* Symbol-table is organized as binary tree's per level.
      Each entry in display-array points to unbalanced binary tree,
      array-index is level of current procedure/function       *)

   display : array [disprange] of identptr;

   (* Error messages *)

   errinx : 0 .. 10 ;       (* Number of errors
                               in current source line *)

   errtot : integer ;       (* total number of errors *)

   (* array to hold error messages from compiler *)

   cmperror : array [1 .. maxermsg] of alphastring ;

   errlist :
             array [ 1 .. maxerlines ] of
               PACKED RECORD
                 pos : 1 .. maxchcnt  ; (* position of error found *)
                 nmr : 1 .. maxermsg   (* number of error found   *)
                END ;

   (* Expression compilation *)

   gattr : attr  ;    (* Describes the expression being compiled *)

   (* Structured Constants *)

   constbegsys,
   simptypebegsys,
   typebegsys,
   blockbegsys,
   selectsys,
   facbegsys,
   statbegsys,
   typedels       :    setofsys ;

   rw  : array [ 0 .. 32  ] of alpha    ; (* Reserved words       *)
   frw : array [ 0 .. 8   ] of 0 .. 33  ; (* no res wrds          *)
   rsy : array [ 0 .. 32  ] of symbol   ; (* Reserved symbols     *)
   ssy : array [ 0 .. 127 ] of symbol   ; (* Not reserved symbols *)
   rop : array [ 0 .. 32  ] of operator ; (* Reserved             *)
   sop : array [ 0 .. 127 ] of operator ;
   na  : array [ 0 .. 20  ] of alpha    ; (* Standard proc/func's *)


  PROCEDURE EndLinelist  ;
  (**********************)

    (* Ends line on listfile, takes care of pages, headings etc *)

    VAR

       i : integer ;


    BEGIN

      listlines := listlines + 1 ;

      (* new page needed ? *)
      IF listlines > maxpage
        THEN
          BEGIN
            (* write formfeed and header *)
            writeln ( listfile, chr(12) ) ;
            writeln ( listfile ) ;
            write ( listfile, 'Pascal-M V1.4 Source listing     ') ;
            FOR i := 1 TO 20 DO
              write ( listfile, filename[i] ) ;
            writeln ( listfile, '               Page ', listpages:4 ) ;
            writeln ( listfile ) ;
            writeln ( listfile ) ;
            listpages := listpages + 1 ;
            listlines := 3
          END ;

    END ; (* EndLineList *)


  PROCEDURE BeginLine ;
  (*******************)

  VAR

    kar : char ;

  (* reads line from source into line *)

    BEGIN

     IF eof(infile)
       THEN
          BEGIN
            writeln ;
            writeln ('Premature END of source file') ;
            EndLineList ;
            writeln(listfile) ;
            EndLineList ;
            writeln (listfile,'Premature END of source file') ;
            writeln(outfile,'P1010000');
            halt ;
          END
       ELSE
         BEGIN
           linelength := 0 ;
           linecount := linecount + 1 ;
           EndLineList ;
           write ( listfile, ' ', linecount:4, ': ' ) ;
           WHILE ( NOT eoln(infile) ) AND
                 ( linelength < maxchcnt ) DO
             BEGIN
               linelength := linelength + 1 ;
               {
               IF status(infile) <> 0
                 THEN
                   BEGIN
                     writeln ;
                     writeln ('Premature END of source file') ;
                     EndLineList ;
                     writeln(listfile) ;
                     EndLineList ;
                     writeln (listfile,'Premature END of source file') ;
                     writeln(outfile,'P1010000');
                     linelength := $EXIT
                   END ;
               }
               read ( infile, kar) ;
               IF kar = chr(aform)
                 then
                   write (listfile, ' ')
                 else
                   write(listfile, kar)  ;
               line[linelength] := kar ;
             END ;
           readln(infile ) ;
           writeln ( listfile ) ;
           chcnt := 0 ;
           linpos := 0
         END ;

    END ; (* BeginLine *)


  PROCEDURE EndLine ;
  (*****************)

  (* writes line to listing-file and reports errors if any
     on terminal and in listing *)

   VAR
       lastpos,
       freepos,
       currpos,
       currnmr,
       f, k     : integer ;

   BEGIN

     (* Output Error-messages *)

     IF errinx > 0
       THEN
         BEGIN
           writeln ;
           EndLineList ;
           write ( ' ', linecount:4, ': ' ) ;
           FOR k := 1 to linelength DO
             write ( line [ k ] ) ;
           writeln ;
           write(' **** ') ;
           write( listfile, ' **** ') ;
           lastpos := 0 ;
           freepos := 1 ;
           FOR k :=1 to errinx DO
             BEGIN
               currpos := errlist[k].pos ;
               currnmr := errlist[k].nmr ;
               IF currpos <= lastpos
                 THEN
                   BEGIN
                     write (',')  ;
                     write (listfile, ',')
                   END
                 ELSE
                   BEGIN
                     WHILE freepos < currpos DO
                       BEGIN
                         write (' ') ;
                         write (listfile, ' ') ;
                         freepos := freepos + 1 ;
                       END ;
                     write ('^') ;
                     write (listfile, '^') ;
                     lastpos := currpos ;
                   END ;
                 IF currnmr < 10
                   THEN
                     f := 1
                   ELSE
                     IF currnmr < 100
                       THEN
                         f := 2
                       ELSE
                         f := 3 ;
                 write (currnmr:f) ;
                 write (listfile, currnmr:f) ;
                 freepos := freepos + f + 1 ;
             END ;
           writeln ( listfile ) ;
           writeln ;

           (* Display the meaning of the error-numbers *)

           FOR k := 1 to errinx DO
             BEGIN
               currnmr := errlist[k].nmr ;
               EndLineList ;
               writeln ( ' ':11, currnmr:3, ' = ',cmperror[currnmr] ) ;
               writeln ( listfile, ' ':13, currnmr:3, ' = ',cmperror[currnmr])
             END ;
           errinx := 0 ;

         END

     END ; (* EndLine *)

  PROCEDURE Error ( ferrnr : integer ) ;
  (************************************)

  (* adds error to errors for current line *)

    BEGIN

      errflag := true ;
      errtot := errtot + 1 ;
      IF errinx >= 9
        THEN
          BEGIN
            errlist[10].nmr := maxermsg ;
            errinx := 10
          END
        ELSE
          BEGIN
            errinx := errinx + 1 ;
            errlist[errinx].nmr := ferrnr
          END ;
      errlist [errinx].pos := linpos

    END ; (* Error *)

    FUNCTION StringSize ( lvp : constptr ) : integer ;
    (************************************************)

    (* calculates length of string, build up of linked list
       of constptr's cells, lvp points at list-head  *)

      VAR
           n : integer ; (* length counter *)

      BEGIN

        n := 0 ;

        (* walk through linked list until end *)
        REPEAT
          n := n + lvp^.slgth ;
          lvp := lvp^.sptr       (* next cell *)
        UNTIL lvp = nil ;

        StringSize := n          (* return result *)

    END ; (* StringSize *)

  PROCEDURE InSymbol ;
  (******************)

  (* Read next basic symbol of source program and return its
     description in the global variables
     sy   = symbol found
     op   = operator if any
     id   = name if identifier
     val  = value if constant
     lgth = number of characters if string

     Symbols parsed :

     - identifier, reserved or declared
     - integer-constant
     - string-constant
     - colon or becomes
     - period or double period (=colon)
     - operator-symbols less, less or equal, greater, greater or equal,
                      not equal
     - hex-constant, returned as integer-constant
     - single character symbols as found in table ssy

     - Comments, new lines  skipped, illegal symbols and
       characters flagged                                  *)

      VAR
          i, k : integer ;
          chstring : strng ;
          lvp, lvp1 : constptr ;
          test, comnt : boolean ;
          quotechar : char ;

      PROCEDURE NextChar ;
      (******************)

      (* returns next character from input in
           ch  = character,
           ich = ordinal value of character,
         translates tab in one space
         keeps count of position in line and real incl tabs *)


        BEGIN

          (* new line from input necessary ? *)
          IF chcnt > linelength
            THEN
              BEGIN
                EndLine ;
                BeginLine
              END ;
         chcnt := chcnt + 1 ;
         linpos := linpos + 1 ;
         IF chcnt > linelength
           THEN
             ch := ' '     (* return blank if end of line reached *)
           ELSE
             ch := line [ chcnt ] ;

        (* get ordinal value *)
        ich := ord(ch) ;

        (* correct for form-feed and tab-characters *)
        IF ich = atab
          THEN
            BEGIN
              ch := ' ' ;
              ich := ord(ch) ;
              linpos := linpos + 7
            END
        ELSE IF ich = aform
          THEN
            BEGIN
              listlines := maxpage + 1 ;
              ch := ' ' ;
              ich := ord(ch)
            END ;

      END ; (* NextChar *)

  BEGIN  (* InSymbol *)
  (*******************)

    REPEAT (* until symbol found, skip comments *)

     comnt := true ;

     (* skip blanks *)
     WHILE ch = ' ' DO
       NextChar ;


     IF ( ( ch >= 'A') and ( ch <= 'Z') or
          ( ch >= 'a') and ( ch <= 'z') )
       THEN
         (* identifier found, starts with upper- or lowercase :

                           --------
          ident -->-------! letter !----------->--
                     ^     --------      v
                     !     --------      !
                      --<-! letter !-----
                     !     --------      !
                     !     --------      v
                      --<-! digit  !-----
                          --------             *)

         BEGIN
           kk := 0 ;
           id := '        ' ;
           REPEAT
             (* only 8 characters significant in name, skip rest *)
             IF kk < 8
               THEN
                 BEGIN
                   kk := kk + 1 ;
                   (* translate uppercase to lowercase in identifier *)
                   IF ( ch >= 'A') and ( ch <= 'Z')
                     THEN
                       BEGIN
                         ich := ich + 32 ;
                         ch := chr (ich )
                       END ;
                   id [kk]  := ch
                 END ;
             NextChar
           UNTIL not ( (( ch >= '0' ) and ( ch <= '9' )) or
                       (( ch >= 'a' ) and ( ch <= 'z' )) or
                       (( ch >= 'A' ) and ( ch <= 'Z')) ) ;
           sy := ident ;
           op := noop  ;
           (* search for reserved name, fast search by using
              length as index with table kk                  *)
           i := frw [kk-1] ;

           WHILE i <= frw [kk] -1 DO
             BEGIN
               IF rw[i] = id
                 THEN
                   BEGIN
                     sy := rsy[i] ;
                     op := rop[i]
                   END ;
               i := i + 1
             END
         END

       ELSE IF ( ch >= '0' ) and ( ch <= '9' )
         THEN
             (* Integer constant, only digits expected :

                               --------
      integer-constant ->-----! digit  !----------->--
                         ^     --------      v
                         !     --------      !
                          --<-! digit  !-----
                               --------             *)

           BEGIN
             op := noop ;
             test := false ;
             k := 0 ;
             REPEAT
               IF k <= mxintio
                 THEN
                   k := k * 10 + ich - ord('0')
                 ELSE
                   test := true ;
               NextChar
             UNTIL ( ch < '0' ) or ( ch > '9' ) ;
             IF test
               THEN
                 Error(171) ; (* number too large *)
             val.ival := k ;
             sy := intconst
           END

       ELSE IF ( ch = '''' ) or ( ch = '"')
         THEN
           (* String, starts with and ends with (d)quote *)
           (* Put string from input in linked list of
              constants, each constant containing max 8 char's *)
           BEGIN
             lgth := 0 ;
             sy := stringconst ;
             op := noop ;
             lvp := nil ;
             quotechar := ch ; (* start and end with same quote *)
             REPEAT
               REPEAT
                 IF ( chcnt > linelength )
                   THEN
                     Error(170) ; (* string must be on one line *)
                 NextChar ;
                 lgth := lgth + 1 ;
                 IF lgth > strglgth
                   THEN
                     BEGIN
                       new (lvp1 ) ;
                       IF lvp <> nil
                         THEN
                           lvp^.sptr := lvp1  (* link to list *)
                         ELSE
                           val.valp := lvp1 ; (* new list *)
                       lvp := lvp1 ;
                       lvp^.sval := chstring ;
                       lvp^.slgth := strglgth ;
                       lvp^.sptr := nil ;
                       lgth := 1
                     END ;
                 chstring [ lgth ] := ich
               UNTIL ch = quotechar ;
               NextChar
             UNTIL ch <> quotechar ;
             lgth := lgth - 1 ;      (* nr of char's in string *)
             IF ( lgth = 1 ) and ( lvp = nil )
               THEN
                 val.ival := chstring[1] (* only one character *)
               ELSE
                 (* create list-head pointer describing string *)
                 IF lgth <> 0
                   THEN
                     BEGIN
                       new ( lvp1 ) ;
                       IF lvp <> nil
                         THEN
                           lvp^.sptr := lvp1
                         ELSE
                           val.valp := lvp1 ;
                       lvp1^.sval := chstring ;
                       lvp1^.slgth := lgth  ;
                       lvp1^.sptr := nil ;
                       lgth := StringSize(val.valp) ;
                     END
           END

       ELSE IF ch = ':'
         THEN
           BEGIN
             op := noop ;
             NextChar ;
             IF ch = '='
               THEN
                 BEGIN
                   sy := becomes ;
                   NextChar
                 END
               ELSE
                 sy := colon
             END

       ELSE IF ch = '.'
         THEN
           BEGIN
             op := noop ;
             NextChar ;
             IF ch = '.'
               THEN
                 BEGIN
                   sy := colon ;
                   NextChar
                 END
               ELSE
                 sy := period
             END

       ELSE IF ch = '<'
         THEN
           BEGIN
             NextChar ;
             sy := relop ;
             IF ch = '='
               THEN
                 BEGIN
                   op := leop ;
                   NextChar
                 END
               ELSE
                 IF ch = '>'
                   THEN
                     BEGIN
                       op := neop ;
                       NextChar
                     END
                   ELSE
                     op := ltop
            END

       ELSE IF ch = '>'
         THEN
           BEGIN
             NextChar ;
             sy := relop ;
             IF ch = '='
               THEN
                 BEGIN
                   op := geop ;
                   NextChar
                 END
               ELSE
                 op := gtop
           END

       ELSE IF ch = '('
         THEN
           BEGIN
             NextChar ;
             IF ch = '*' (* skip comment *)
               THEN
                 BEGIN
                   NextChar ;
                   REPEAT
                     WHILE ch <> '*' DO
                       NextChar ;
                     NextChar ;
                   UNTIL ch = ')' ;
                   NextChar ;
                   comnt := false
                 END
               ELSE
                 BEGIN
                   sy := lparent ;
                   op := noop
                 END
           END

       ELSE IF ch = '{' (* comment *)
         THEN
           BEGIN
             (* Skip until right acculade found *)
             WHILE ch <> '}' DO
               NextChar ;
             NextChar ;
             comnt := false
           END


       ELSE IF ch = '$' (* hex constant, 0 .. 9, A .. F *)
         THEN
           BEGIN
             op := noop ;
             k := 0 ;
             REPEAT
               NextChar ;
               IF ( ch >= '0' ) and ( ch <= '9' )
                 THEN
                   i := ich - ord('0')
               ELSE IF ( ch >= 'A') and ( ch <= 'F')
                 THEN
                   i := ich - ord('A') + 10
               ELSE IF ( ch >= 'a') and ( ch <= 'f')
                 THEN
                   i := ich - ord('a') + 10
               ELSE
                       i := -1 ;
               IF i >= 0
                 THEN
                   k := 16*k + i ;
             UNTIL i < 0 ;
             val.ival := k ;
             sy := intconst
           END

       ELSE  (* other single character symbols *)
         BEGIN
           sy := ssy[ich] ;
           op := sop[ich] ;
           NextChar ;
           IF sy = othersy
             THEN
               Error(176)   (* illegal character found *)
         END ;

    UNTIL comnt ;

(* debug
write ('Symbol found = ') ;

CASE sy OF

ident : writeln ( 'ident ' ) ;
intconst : writeln ( 'intconst ' ) ;
realconst : writeln ( 'realconst ' ) ;
stringconst : writeln ( 'stringconst ' ) ;
notsy : writeln ( 'notsy ' ) ;
mulop : writeln ( 'mulop ' ) ;
addop : writeln ( 'addop ' ) ;
relop : writeln ( 'relop ' ) ;
lparent : writeln ( 'lparent ' ) ;
rparent : writeln ( 'rparent ' ) ;
lbrack : writeln ( 'lbrack ' ) ;
rbrack : writeln ( 'rbrack ' ) ;
comma : writeln ( 'comma ' ) ;
semicolon : writeln ( 'semicolon ' ) ;
period : writeln ( 'period ' ) ;
colon : writeln ( 'colon ' ) ;
becomes : writeln ( 'becomes ' ) ;
constsy : writeln ( 'constsy ' ) ;
typesy : writeln ( 'typesy ' ) ;
varsy : writeln ( 'varsy ' ) ;
funcsy : writeln ( 'funcsy ' ) ;
progsy : writeln ( 'progsy ' ) ;
procsy : writeln ( 'procsy ' ) ;
setsy : writeln ( 'setsy ' ) ;
packedsy : writeln ( 'packedsy ' ) ;
arraysy : writeln ( 'arraysy ' ) ;
recordsy : writeln ( 'recordsy ' ) ;
externsy : writeln ( 'externsy ' ) ;
forwardsy : writeln ( 'forwardsy ' ) ;
beginsy : writeln ( 'beginsy ' ) ;
ifsy : writeln ( 'ifsy ' ) ;
casesy : writeln ( 'casesy ' ) ;
repeatsy : writeln ( 'repeatsy ' ) ;
whilesy : writeln ( 'whilesy ' ) ;
forsy : writeln ( 'forsy ' ) ;
endsy : writeln ( 'endsy ' ) ;
elsesy : writeln ( 'elsesy ' ) ;
untilsy : writeln ( 'untilsy ' ) ;
ofsy : writeln ( 'ofsy ' ) ;
dosy : writeln ( 'dosy ' ) ;
tosy : writeln ( 'tosy ' ) ;
downtosy : writeln ( 'downtosy ' ) ;
thensy : writeln ( 'thensy ' ) ;
othersy : writeln ( 'othersy ' ) ;
otherwise writeln('unknown symbol found') ;
end;

write('Operator = ') ;
CASE op OF
mul : writeln('mul');
andop : writeln('andop ');
idiv : writeln('idiv ');
imod : writeln( 'imod ');
plus : writeln( 'plus ');
minus : writeln( 'minus ');
orop : writeln( ' orop ');
ltop : writeln( 'ltop ');
leop : writeln( 'leop ');
geop : writeln( 'geop ');
gtop : writeln( 'gtop ');
neop : writeln( 'neop ');
eqop : writeln( 'eqop ');
inop : writeln( 'inop ');
noop : writeln( 'noop ');
otherwise writeln('unknown operator') ;

  END;

write  ( 'id = ' ) ;
for i := 1 to 8 DO
 write( id [i] ) ;
writeln ;
  debug *)

  END ; (* InSymbol *)

  PROCEDURE Enterid ( fcp : identptr) ;
  (***********************************)

  (* Enter ID pointed at by FCP into the name-table,
     which on each declaration level is organized as
     an unbalanced binary tree                       *)

  var

    nam : alpha ;
    lcp, lcp1 : identptr;
    lleft : boolean ;

  BEGIN

    nam := fcp^.name ;
    lcp := display[top] ;
    IF lcp = nil
      THEN
        (* empty tree , enter at root *)
        display[top] := fcp
      ELSE
        BEGIN
          (* find place in tree where to put identifier *)
          REPEAT
            lcp1 := lcp ;
            (* Decide if left or right to go, if
               name conflict follow right link *)
            IF lcp^.name = nam
              THEN
                BEGIN
                  (* already present *)
                  Error(101) ;
                  lcp := lcp^.rlink ;
                  lleft := false
                END
            ELSE IF lcp^.name < nam
              THEN
                BEGIN
                  (* go right *)
                  lcp := lcp^.rlink ;
                  lleft := false
                END
            ELSE
              BEGIN
                (* go left *)
                lcp := lcp^.llink ;
                lleft := true
              END
          UNTIL lcp = nil ;

          (* place found, link identifier in tree *)
          IF lleft
            THEN
              lcp1^.llink := fcp
            ELSE
              lcp1^.rlink := fcp
        END ;
    (* identifier is new leave, no left or right pointers yet *)
    fcp^.llink := nil ;
    fcp^.rlink := nil

  END ;  (* Enterid *)


  PROCEDURE SearchSection ( fcp : identptr; var fcp1 : identptr) ;
  (**************************************************************)

  (* To find record fields and forward declared
     procedure id's
     called by
        - procedure ProcedureDeclaration
        - procedure Selector
     search starts in tree at identifier pointed by fcp,
     searched for identifier with same name as id        *)

  BEGIN

    (* indicate not found yet *)
    fcp1 := nil ;
    WHILE ( fcp <> nil ) and ( fcp1 = nil ) DO
      IF fcp^.name = id
       THEN
         (* found, stop search and return pointer *)
         fcp1 := fcp
       ELSE
         (* not found, descend down right or left *)
         IF fcp^.name < id
           THEN
             fcp := fcp^.rlink
           ELSE
             fcp := fcp^.llink

  END ; (* SearchSection *)


  PROCEDURE Searchid ( fidcls : setofids ; var fcp : identptr) ;
  (************************************************************)

  (* Search identifier in symbol-table, start at display-level
     top and go down to bottom display. Search based on name id.
     If found then type is checked with types in fidcls.
     If not found then return unknown type-description *)

  VAR

    lcp : identptr;

  BEGIN

  (* search starts at top and identifier is not found yet *)
  disx := top + 1 ;
  fcp := nil ;

  WHILE ( disx > 0 ) and ( fcp = nil ) DO
    BEGIN
      disx := disx - 1 ;

      (* new root *)
      lcp := display [ disx ] ;

      (* search tree *)
      WHILE ( lcp <> nil ) and ( fcp = nil ) DO
        (* decide if found or go down right or left *)
        IF lcp^.name = id
          THEN
            (* check if type is in fidcls *)
            IF lcp^.klass in fidcls
              THEN
                fcp := lcp
              ELSE
                BEGIN
                  (* not appropiate class *)
                  IF prterr
                    THEN
                      Error(103) ;
                  (* go right, maybe double declaration *)
                  lcp := lcp^.rlink
                END
        ELSE IF lcp^.name < id
          THEN
            (* go right *)
            lcp := lcp^.rlink
        ELSE
          (* go left *)
          lcp := lcp^.llink

    END ;

    (* Search not successful, suppress Error message
       in case of forward referenced type id in pointer type
       definition --> procedure SimpleType                  *)

    IF prterr and ( fcp = nil )
      THEN
        BEGIN
          Error(104) ;
          (* To avoid returning nil, reference an entry
             for undeclared ID of appropiate class
             ---> PROCEDURE Enterundecl                     *)
          IF types in fidcls
            THEN
              fcp := utypptr
          ELSE IF vars in fidcls
            THEN
              fcp := uvarptr
          ELSE IF field in fidcls
            THEN
              fcp := ufldptr
          ELSE IF konst in fidcls
            THEN
              fcp := ucsptr
          ELSE IF proc in fidcls
            THEN
              fcp := uprcptr
          ELSE
            fcp := ufctptr ;
        END

  END ; (* Searchid *)

  PROCEDURE GetBounds ( fsp : structptr ; VAR fmin, fmax : integer ) ;
  (******************************************************************)

  (* Get internal bounds of subrange or scaler type ,
     Assume FSP <> nil and fsp^.form <= subrange and fsp <> intptr,
     returns bounds in fmin and fmax *)

  BEGIN
    IF fsp^.form = subrange
      THEN
        BEGIN
          fmin := fsp^.min.ival ;
          fmax := fsp^.max.ival
        END
      ELSE   (* scalar *)
        BEGIN
          fmin := 0 ;
          IF fsp = charptr
            THEN
              (* characters *)
              fmax := 127
            ELSE
              (* other scalars *)
              fmax := 0
        END

  END ; (* GetBounds *)

  PROCEDURE HexOut ( word : integer ) ;
  (***********************************)

  (* write word as four hex characters *)

  VAR

    k : integer ;

  BEGIN

    sumcheck := sumcheck + word ;
    WHILE word < 0 DO
      word := word + 256 ;
    k := word div 16 ;
    word := word mod 16 ;
    IF k < 10
      THEN
        write ( outfile, chr (ord ('0') + k) )
      ELSE
        write ( outfile, chr (ord ('A') + k - 10 ) ) ;
    IF word < 10
      THEN
        write ( outfile, chr (ord ('0') + word) )
      ELSE
        write ( outfile, chr (ord ('A') + word  - 10 ) ) ;

  END ; (* HexOut *)

  PROCEDURE WriteOut ;
  (******************)

  (* write P1-record with code buffer and checksum,
    codebuffer empty and instruction-counter is updated *)

  VAR

    i : integer ;

    BEGIN

      IF icn <> 0
        THEN
          BEGIN
            write( outfile, 'P1' ) ;
            sumcheck := 0 ;
            HexOut (icn + 1 ) ;
            for i := 0 to icn - 1 DO
              HexOut ( codebuf [i] ) ;
            HexOut ((16383 - sumcheck ) mod 256 ) ;
            writeln ( outfile )
          END ;
      ic := ic + icn ;
      icn := 0

  END ; (* WriteOut *)


  PROCEDURE Block ( fsys : setofsys ; fsy : symbol ; fprocp : identptr);
  (********************************************************************)

  (* compiles block *)

  TYPE

    oprange = 0 .. 255 ;

  VAR

    lsy : symbol ;
    test : boolean ;

    PROCEDURE Bytegen  ( byte : integer ) ;
    (*************************************)

    (* add byte to codebuffer, empty codebuffer if necessary *)

    BEGIN

      IF icn = maxcode + 1
        THEN
          WriteOut ;
      codebuf [icn] := byte ;
      icn := icn + 1

    END ; (* Bytegen *)

    PROCEDURE Addrgen ( addr : integer ) ;
    (************************************)

    (* put addr as two bytes in codebuffer *)

    BEGIN

      Bytegen ( addr div 256 ) ;
      Bytegen ( addr mod 256 )

    END ; (* Addrgen *)

    PROCEDURE GenUJPent ( fop : oprange ; word : integer ) ;
    (******************************************************)

    (* generate fop instruction, followed by fp2 as word *)

    BEGIN

      Bytegen ( fop ) ;
      Addrgen ( word)

    END ; (* GenUJPent *)

    PROCEDURE PlantAddress ( location, value : integer ) ;
    (****************************************************)

    (* plant value at location in code ,
       generate P2-record if loc not in codebuffer *)

    BEGIN

      IF value <> 0
        THEN
          IF ( location >= ic ) and ( location - ic < icn )
            THEN
              BEGIN
                (* place in codebuffer *)
                (* correct for offset in codebuffer *)
                location := location - ic ;
                codebuf [location] := value div 256 ;
                codebuf [location + 1] := value mod 256
              END
            ELSE
              BEGIN
                (* Generate P2-record with location and value *)
                WriteOut ;
                write (outfile, 'P2') ;
                sumcheck := 0 ;
                (* first location *)
                HexOut ( location div 256 ) ;
                HexOut ( location mod 256 ) ;
                (* next value to be planted *)
                HexOut ( value div 256 ) ;
                HexOut ( value mod 256 ) ;
                HexOut ((16383 - sumcheck ) mod 256 ) ;
                writeln (outfile )
              END ;

    END ; (* PlantAddress *)

    PROCEDURE Skip ( fsys : setofsys ) ;
    (**********************************)

    (* Skip input chstring until relevant symbol found *)

    BEGIN

      WHILE ( not ( sy in fsys ) ) DO
        InSymbol

    END ; (* Skip *)

    PROCEDURE Test1 ( fsys : setofsys ; errornr : integer ) ;
    (*******************************************************)

    (* Test if current symbol sy in set fsys, if not generate
       error with number errornr and skip until sy in fsys *)

    BEGIN

      IF not ( sy in fsys)
        THEN
          BEGIN
            Error(errornr) ;
            Skip (fsys)
          END

    END ; (* Test1 *)

    PROCEDURE Test2 (fsys : setofsys; errornr : integer; s2 : setofsys);
    (******************************************************************)

    (* Test if current symbol sy in set fsys, if not generate
       error with number errornr and skip until sy in fsys + s2 *)

    BEGIN

      IF not ( sy in fsys)
        THEN
          BEGIN
            Error(errornr) ;
            Skip (fsys + s2 )
          END

    END ; (* Test2 *)

    PROCEDURE Intest ( sym : symbol ; errornr : integer ) ;
    (*****************************************************)

    (* Test if current symbol sy is sym, if not generate
       error with number errornr , get next symbol from input *)

    BEGIN

      IF sy = sym
        THEN
          InSymbol
        ELSE
          Error(errornr)

    END ; (* Intest *)

    PROCEDURE Constant ( fsys : setofsys    ;
    (****************)   var fsp : structptr      ;
                         var fvalu : valu ) ;

    (* checks if constant, declared or string-constant or integer,
       returns decsription in fsp , value in fvalu *)

    TYPE

      signs = ( none, pos, neg ) ;

    VAR

       lsp  : structptr ;
       lcp  : identptr;
       sign : signs ;

    BEGIN

      (* not a constant, value zero *)
      lsp := nil ;
      fvalu.ival := 0 ;

      Test2 (constbegsys, 50, fsys ) ;
      IF sy in constbegsys
        THEN
          BEGIN
            IF sy = stringconst
              THEN
                BEGIN
                  IF lgth = 1
                    THEN
                      lsp := charptr
                    ELSE
                      BEGIN
                        new (lsp) ;
                        lsp^.stype := charptr ;
                        lsp^.indextype := nil ;
                        lsp^.size := StringSize(val.valp) ;
                        lsp^.form := arrays ;
                      END ;
                  fvalu := val ;
                  InSymbol
                END
              ELSE
                (* other than string-constant, first check if sign *)
                BEGIN
                  sign := none ;
                  IF ( sy = addop ) and ( op in [ plus, minus ] )
                    THEN
                      BEGIN
                        IF op = plus
                          THEN
                            sign := pos
                          ELSE
                            sign := neg ;
                        InSymbol
                      END ;

                  (* constant can be declared identifier or declared constant *)
                  IF sy = ident
                    THEN
                      BEGIN
                        Searchid ( [konst],lcp ) ;
                        lsp := lcp^.idtype ;
                        fvalu := lcp^.values ;
                        IF sign <> none
                          THEN
                            IF lsp = intptr
                              THEN
                                BEGIN
                                  IF sign = neg
                                    THEN
                                      fvalu.ival := - fvalu.ival ;
                                END
                              ELSE
                                (* sign not allowed *)
                                Error (105) ;
                        InSymbol
                      END
                  ELSE IF sy = intconst
                    THEN
                      BEGIN
                        IF sign = neg
                          THEN
                            val.ival := -val.ival ;
                        lsp := intptr ;
                        fvalu := val ;
                        InSymbol
                      END
                  ELSE
                    BEGIN
                      (* number expected *)
                      Error (106) ;
                      Skip (fsys)
                    END
                END ;

            Test1(fsys,6)

          END ;

      fsp := lsp

    END ; (* Constant *)

    FUNCTION CompTypes ( fsp1 , fsp2 : structptr ) : boolean ;
    (**************************************************)

    (* Decide whether structures pointed at by FSP1 and
       FSP2 are compatible                              *)

    BEGIN

      IF fsp1 = fsp2
        THEN
          CompTypes := true
        ELSE
          IF ( fsp1 <> nil ) and ( fsp2 <> nil )
            THEN
              BEGIN
                IF fsp1^.form = fsp2^.form
                  THEN
                    CASE fsp1^.form OF
                      scalar , records          :
                        CompTypes := false ;
                      subrange, pointer, power  :
                         CompTypes :=
                             CompTypes (fsp1^.stype,fsp2^.stype);
                      arrays                    :
                         CompTypes :=
                             CompTypes (fsp1^.stype,fsp2^.stype)
                              and
                             ( fsp1^.size = fsp2^.size )
                      END

                  (* fsp1^.form <> fsp2^.form ) *)

                  ELSE IF fsp1^.form = subrange
                    THEN
                      CompTypes := CompTypes ( fsp1^.stype, fsp2 )
                  ELSE IF fsp2^.form = subrange
                     THEN
                       CompTypes := CompTypes ( fsp1, fsp2^.stype )
                  ELSE
                    CompTypes := false
            END
          ELSE
              CompTypes := true

    END ; (* CompTypes *)


    FUNCTION IsString ( fsp : structptr ) : boolean ;
    (***********************************************)

    (* Tests if structure pointed by with fsp is string *)

    BEGIN

      IsString := false ;
      IF fsp <> nil
        THEN
          IF fsp^.form = arrays
            THEN
              IF CompTypes (fsp^.stype, charptr )
                THEN
                  IsString := true

    END ; (* IsString *)


    PROCEDURE Typ ( fsys : setofsys ;
    (*************)  var fsp : structptr ;
                    var fsize : addrrange  ) ;

    VAR

      lsp, lsp1, lsp2 : structptr ;
      oldtop : disprange ;
      lcp : identptr;
      lsize, displ : addrrange ;
      lmin, lmax : integer ;

    PROCEDURE simpletype ( fsys : setofsys ;
    (******************)     var fsp : structptr ;
                             var fsize : addrrange ) ;

    (* parses and returns description and size of simple type

       structure parsed :

                          -----------------
       --------->--------! type identifier !--------------->
             !            -----------------            !
             v                                         ^
             !       ---     ------------     ---      !
             !-->---( ( )---! identifier !---( ) )-----!
             !       ---  !  ------------     ---      !
             !            !              !             !
             !            !      ---     !             !
             !             -----( , )----              !
             !                   ---                   !
             !                                         !
             !     ----------    ----    ----------    !
              ----! constant !--( .. )--! constant !---
                   ----------    ----    ----------           *)

    VAR

      lsp, lsp1 : structptr ;
      lcp, lcp1 : identptr;
      ttop : disprange ;
      lcnt : integer ;          (* keeps track of enumerated constants *)
      lvalu : valu ;

    BEGIN

      fsize := 1 ;
      Test2 ( simptypebegsys, 1, fsys ) ;
      IF not ( sy in simptypebegsys )
        THEN
          fsp := nil
        ELSE
          BEGIN
            IF sy = lparent
              THEN
                (* enumeration type *)
                BEGIN
                  ttop := top ;
                  top := savetop ;
                  new ( lsp ) ;
                  lsp^.form := scalar ;
                  lcp1 := nil ;
                  lcnt := 0 ;
                  REPEAT
                    InSymbol ;
                    (* enumeration must be list of constant idents *)
                    IF sy <> ident
                      THEN
                        Error(2)
                      ELSE
                        BEGIN
                          new ( lcp ) ;
                          lcp^.name := id ;
                          lcp^.idtype := lsp ;
                          lcp^.next := lcp1 ;
                          lcp^.values.ival := lcnt ;
                          lcp^.klass := konst ;
                          Enterid ( lcp ) ;
                          lcnt := lcnt + 1 ;
                          lcp1 := lcp ;
                          InSymbol ;
                        END ;
                    Test1 ( fsys+ [comma,rparent],6 ) ;
                  UNTIL sy <> comma ;
                  lsp^.fstfld := lcp1 ;
                  top := ttop ;
                  (* check if enumeration-type fits in one byte *)
                  IF lcnt < 257
                    THEN
                      lsp^.size := 1
                    ELSE
                      lsp^.size := intsize ;
                  Intest ( rparent, 4 )
                END

            ELSE IF sy = ident
              THEN
                (* type-identifier or declared constant *)
                BEGIN
                  Searchid ( [types,konst], lcp ) ;
                  InSymbol ;
                  IF lcp^.klass = types
                    THEN
                      lsp := lcp^.idtype
                    ELSE
                      (* subrange starts with constant *)
                      BEGIN
                        new ( lsp ) ;
                        lsp^.stype := lcp^.idtype ;
                        lsp^.form := subrange ;
                        IF IsString ( lsp^.stype)
                          THEN
                            BEGIN
                              Error ( 148 ) ;
                              lsp^.stype := nil
                            END ;
                        lsp^.min := lcp^.values ;
                        lsp^.size  := lcp^.idtype^.size ;
                        Intest ( colon, 5 ) ;
                        Constant ( fsys, lsp1, lvalu ) ;
                        lsp^.max := lvalu ;
                        IF lsp^.stype <> lsp1
                          THEN
                            Error (107)
                      END
                  END (* sy is ident *)

             ELSE (* must be subrange starting with a constant *)
               BEGIN
                 new ( lsp ) ;
                 lsp^.form := subrange ;
                 Constant ( fsys + [colon], lsp1, lvalu ) ;
                 IF IsString ( lsp1)
                   THEN
                     BEGIN
                       Error ( 148 ) ;
                       lsp1 := nil
                     END ;
                  lsp^.stype := lsp1 ;
                  lsp^.min := lvalu ;
                  lsp^.size := intsize ;
                  Intest ( colon, 5 ) ;
                  Constant ( fsys, lsp1 , lvalu ) ;
                  lsp^.max := lvalu ;
                  IF lsp^.stype <> lsp1
                    THEN
                      Error ( 107 ) ;
                 END ;
           IF lsp <> nil
             THEN
               IF lsp^.form = subrange
                 THEN
                   IF lsp^.min.ival > lsp^.max.ival
                     THEN
                       (* startvalue of subrange must be smaller
                          than endvalue *)
                       Error (102) ;
            fsp := lsp ;
            Test1 ( fsys, 6 )
          END ;
      IF lsp <> nil
        THEN
          fsize := lsp^.size ;

    END ; (* simpletype *)


    PROCEDURE fieldlist ( fsys : setofsys ) ;
    (***************************************)

    (* Parses field-list,
       returns displ = total space needed to store field-list in record
       uses nxt^ as linked list, see Jensen & wirth page 116
       for syntax diagram *)

    VAR

      lcp , lcp1, nxt, nxt1 : identptr;
      lsp, lsp3 : structptr ;
      minsize, maxsize, lsize : addrrange ;
      lvalu : valu ;

    BEGIN

      nxt1 := nil ;
      lsp := nil ;
      Test2 ( [ident, casesy], 19, fsys  ) ;
      WHILE sy = ident DO
        BEGIN
          nxt := nxt1 ;
          REPEAT
            IF sy <> ident
              THEN
                Error ( 2 )
              ELSE
                BEGIN
                  new ( lcp ) ;
                  lcp^.name := id ;
                  lcp^.idtype := nil ;
                  (* link to list *)
                  lcp^.next := nxt ;
                  lcp^.klass := field ;
                  nxt := lcp ;
                  Enterid ( lcp ) ;
                  InSymbol
                END ;
            Test2 ( [comma,colon],6,fsys + [semicolon, casesy] ) ;
            test := sy <> comma ;
            IF not test
             THEN
               InSymbol
          UNTIL test ;
          Intest ( colon, 5 ) ;
          Typ ( fsys + [ casesy,semicolon ] , lsp, lsize ) ;
          WHILE nxt <> nxt1 DO
            BEGIN
              nxt^.idtype := lsp ;
              nxt^.vaddr := displ ;
              nxt := nxt^.next ;
              displ := displ + lsize
            END ;
          nxt1 := lcp ;
          IF sy = semicolon
            THEN
              BEGIN
                InSymbol ;
                Test2 ( [ident, casesy, semicolon], 19, fsys )
              END
        END ; (* WHILE *)
      nxt := nil ;
      WHILE nxt1 <> nil DO
        BEGIN
          lcp := nxt1^.next ;
          nxt1^.next := nxt ;
          nxt := nxt1 ;
          nxt1 := lcp
        END ;

      (* field-list may contain a variant part *)
      IF sy = casesy
        THEN
          BEGIN
            InSymbol ;
            IF sy = ident
              THEN
                BEGIN
                  (* suppress error messages if forward ref pointer *)
                  prterr := false ;
                  Searchid ( [ types ] , lcp1 ) ;
                  prterr := true ;
                  IF lcp1 = nil
                    THEN
                      BEGIN
                        new ( lcp ) ;
                        lcp^.name := id ;
                        lcp^.idtype := nil ;
                        lcp^.klass := field ;
                        lcp^.next := nil ;
                        lcp^.vaddr := displ ;
                        Enterid ( lcp ) ;
                        InSymbol ;
                        Intest ( colon, 5 ) ;
                        IF sy = ident
                          THEN
                            BEGIN
                              Searchid (  [ types], lcp1 ) ;
                              lsp1 := lcp1^.idtype ;
                              IF lsp1 <> nil
                                THEN
                                  BEGIN
                                    displ := displ + lsp1^.size ;
                                    IF ( lsp1^.form <= subrange ) or
                                       ( IsString ( lsp1 ) )
                                      THEN
                                        BEGIN
                                          IF IsString ( lsp1 )
                                            THEN
                                              Error ( 177 ) ;
                                          lcp^.idtype := lsp1
                                        END
                                      ELSE
                                        Error ( 110 )
                                  END ;
                              InSymbol
                            END
                          ELSE
                            BEGIN
                              Error ( 2 ) ;
                              Skip ( fsys + [ ofsy, lparent ] )
                            END
                      END
                    ELSE
                      BEGIN
                        InSymbol ;
                        lsp1 := lcp1^.idtype
                      END

                END
              ELSE
                BEGIN
                  Error ( 2 ) ;
                  Skip ( fsys+ [ofsy,lparent] )
                END ;

            Intest ( ofsy, 8 ) ;
            minsize := displ ;
            maxsize := displ ;
            REPEAT
              lsp2 := nil ;
              REPEAT
                Constant ( fsys +[comma,colon,lparent],lsp3,lvalu ) ;
                IF not CompTypes ( lsp1, lsp3 )
                  THEN
                    Error ( 111 ) ;
                test := sy <> comma ;
                IF not test
                  THEN
                    InSymbol
              UNTIL test ;
              Intest ( colon, 5 ) ;
              Intest ( lparent, 9 ) ;
              fieldlist ( fsys + [rparent,semicolon] ) ;
              IF displ > maxsize
                THEN
                  maxsize := displ ;
              IF sy = rparent
                THEN
                  BEGIN
                    InSymbol ;
                    Test1 ( fsys + [semicolon],6 )
                  END
                ELSE
                  Error ( 4 ) ;
              test := sy <> semicolon ;
              IF not test
                THEN
                  BEGIN
                    displ := minsize ;
                    InSymbol
                  END
            UNTIL test ;
            displ := maxsize ;

          END  (* case symbol *)

      END ; (* fieldlist *)

    BEGIN (* Typ *)
    (*************)

    (* syntax diagram parsed:

              ------------
 ------------! simpletype !-------------->------------------------------>-
    !         ------------                                             !
    v                                                                  ^
    !                     ---                        ------------      !
    !--------------------( ^ )----------->----------! type-ident !->---!
    !                     ---                        ------------      !
    v  --->------                  ------<-------                      ^
    ! !          !                !              !                     !
    ! !  ------  !  -----    ---  !  ----------  !  ---    --    ----  !
    !---(packed)---(array)--( [ )---!simpletype!---( ] )--(of)--!type!-!
    !    ------     -----    ---     ----------     ---    --    ----  !
    v                                                                  ^
    !              ------           ------------                       !
    !-------------(record)---------! field-list !--------->------------!
    !              ------           ------------                       !
    v                                                                  ^
    !              ------           ------------                       !
     -------------( set  )---------! simpletype !--------->------------
                   ------           ------------                         *)

      Test2 ( typebegsys , 10, fsys ) ;
      IF not ( sy in typebegsys )
        THEN
          fsp := nil
        ELSE
          BEGIN
            IF sy in simptypebegsys
              THEN
                simpletype ( fsys, fsp, fsize )
            ELSE IF sy = arrow
              THEN
                BEGIN
                  (* pointer *)
                  new ( lsp ) ;
                  fsp := lsp ;
                  lsp^.stype := nil ;
                  lsp^.size  := ptrsize ;
                  lsp^.form  := pointer ;
                  InSymbol ;
                  IF sy <> ident
                    THEN
                       Error ( 2 )
                    ELSE
                      BEGIN
                        (* suppress error message *)
                        prterr := false ;
                        Searchid ( [types], lcp ) ;
                        prterr := true ;
                        IF lcp = nil
                          THEN
                            (* forward referenced type id *)
                            BEGIN
                              new ( lcp ) ;
                              lcp^.name := id ;
                              lcp^.idtype := lsp ;
                              lcp^.next := fwptr ;
                              lcp^.klass := types ;
                              fwptr := lcp
                             END
                           ELSE
                             lsp^.stype := lcp^.idtype ;
                        InSymbol ;
                      END ;
                END
              ELSE
                BEGIN
                  IF sy = packedsy
                    THEN
                      BEGIN
                        (* skip packed symbol *)
                        InSymbol ;
                        Test2 ( typedels, 10, fsys )
                      END ;

                  CASE sy OF

                     arraysy :         (* array *)

                       BEGIN
                         InSymbol ;
                         (* '[' expected next *)
                         Intest ( lbrack, 11 ) ;
                         lsp1 := nil ;
                         REPEAT
                           new (lsp ) ;
                           lsp^.stype := lsp1 ;
                           lsp^.indextype := nil ;
                           lsp^.form := arrays ;
                           lsp1 := lsp ;
                           simpletype ( fsys + [comma,rbrack,ofsy],
                                             lsp2 , lsize ) ;
                           lsp1^.size := lsize ;
                           IF lsp2 <> nil
                             THEN
                               IF lsp2^.form <= subrange
                                 THEN
                                   BEGIN
                                     IF lsp2 = intptr
                                       THEN
                                         BEGIN
                                           (* indextype must not
                                              be integer *)
                                           Error ( 149 ) ;
                                           lsp2 := nil
                                          END ;
                                     lsp^.indextype := lsp2
                                   END
                                 ELSE
                                   BEGIN
                                     (* indextype must be scalar
                                        or subrange *)
                                     Error ( 113 ) ;
                                     lsp2 := nil
                                   END ;
                               test := sy <> comma ;
                               IF not test
                                 THEN
                                   InSymbol
                         UNTIL test ;

                         (* next '] of ' expected *)
                         Intest ( rbrack, 12 ) ;
                         Intest ( ofsy, 8 ) ;

                         Typ ( fsys, lsp, lsize ) ;

                         (* calculate length of array *)
                         REPEAT
                           lsp2 := lsp1^.stype ;
                           lsp1^.stype := lsp ;
                           IF lsp1^.indextype <> nil
                             THEN
                               BEGIN
                                 GetBounds ( lsp1^.indextype,
                                               lmin, lmax ) ;
                                 lsize := lsize * ( lmax - lmin + 1 ) ;
                                 lsp1^.size := lsize
                               END ;
                            lsp := lsp1 ;
                            lsp1 := lsp2
                          UNTIL lsp1 = nil
                        END ;

                     recordsy :        (* record *)

                       BEGIN
                         InSymbol ;
                         oldtop := top ;
                         IF top < displimit
                           THEN
                             BEGIN
                               top := top + 1 ;
                               display [top] := nil
                             END
                           ELSE
                             Error ( 172 ) ; (* too many nested *)

                         (* get all fields of record as tree *)
                         displ := 0 ;
                         fieldlist (fsys - [semicolon] + [endsy]);
                         new ( lsp ) ;
                         lsp^.fstfld := display[top] ;
                         lsp^.size := displ ;
                         lsp^.form := records ;
                         top := oldtop ;
                         Intest ( endsy, 13 )
                       END ;

                     setsy :           (* set *)

                       BEGIN
                         InSymbol ;
                         (* set of *)
                         Intest (ofsy, 8 ) ;
                         simpletype ( fsys, lsp1, lsize ) ;
                         IF lsp1 <> nil
                           THEN
                             IF lsp1^.form > subrange
                               THEN
                                 BEGIN
                                    (* base-type scalar or subrange *)
                                    Error ( 115 ) ;
                                    lsp1 := nil ;
                                  END
                                ELSE
                                  (* check set limits *)
                                  GetBounds ( lsp1, lmin, lmax ) ;
                                  IF  ( lmin < 0 ) or
                                      ( lmax > setmax )
                                    THEN
                                      error (169) ;
                         new ( lsp ) ;
                         lsp^.stype := lsp1 ;
                         lsp^.size := setsize ;
                         lsp^.form := power

                       END

                    END ; (* case *)

                  (* return description *)
                  fsp := lsp

                END ;

            Test1 ( fsys, 6 )

          END ;

      IF fsp = nil
        THEN
          fsize := 1
        ELSE
          fsize := fsp^.size

    END ; (* Typ *)

    PROCEDURE ConstDeclaration ;
    (**************************)

    (* compiles constant declaration part.
       Syntax diagram checked ( reserved keyword CONST already parsed):

             --------------------<---------------------------------
            !                                                      !
            v                                                      ^
            !  -------          ---      ----------         ---    !
 CONST --->---! ident !--------( = )----! constant !--->---( ; )----->
               -------          ---      ----------         ---    *)

    VAR

      lcp : identptr;
      lsp : structptr ;
      lvalu : valu ;

    BEGIN

      Test2 ( [ident],2, fsys ) ;
      WHILE sy = ident  DO
        BEGIN
          new ( lcp ) ;
          lcp^.name := id ;
          lcp^.idtype := nil ;
          lcp^.next := nil ;
          lcp^.klass := konst ;
          InSymbol ;
          IF ( sy = relop ) and ( op = eqop )
            THEN
              InSymbol
            ELSE
              Error ( 16 ) ;
          Constant ( fsys + [semicolon], lsp, lvalu ) ;
          Enterid ( lcp ) ;
          lcp^.idtype := lsp ;
          lcp^.values := lvalu ;
          IF sy = semicolon
            THEN
              BEGIN
                InSymbol ;
                Test1 ( fsys + [ident], 6 )
              END
            ELSE
              Error ( 14 )
        END

    END ; (* ConstDeclaration *)


    PROCEDURE TypeDeclaration ;
    (*************************)

    (* compiles type-declaration
       Syntax-diagram checked (reserved keyword TYPE already parsed):


           ------------------------<----------------------------
          v                                                     ^
          !    -------           ---      ------------      ---  !
   TYPE ------! ident !---------( = )----! type-ident !----( ; )--->-
           !   -------        !  ---      ------------      ---
           ^                  v
           !  -----    ---    !
            -!ident!--( , )---
              -----    ---                                      *)

    VAR

      lcp , lcp1, lcp2 : identptr;
      lsp : structptr ;
      lsize : addrrange ;

    BEGIN

      Test2 ( [ident], 2 , fsys ) ;
      WHILE sy = ident DO
        BEGIN
          new ( lcp ) ;
          lcp^.name := id ;
          lcp^.idtype := nil ;
          lcp^.klass := types ;
          InSymbol ;
          IF ( sy = relop ) and ( op = eqop )
            THEN
              InSymbol
            ELSE
              Error ( 16 ) ;
          Typ ( fsys + [ semicolon ], lsp, lsize ) ;
          Enterid ( lcp ) ;
          lcp^.idtype := lsp ;

          (* Has any forward reference been satisfied *)

          lcp1 := fwptr ;
          WHILE lcp1 <> nil DO
            BEGIN
              IF lcp1^.name = lcp^.name
                THEN
                  BEGIN
                    lcp1^.idtype^.stype := lcp^.idtype ;
                    IF lcp1 <> fwptr
                      THEN
                        lcp2^.next := lcp1^.next
                      ELSE
                        fwptr := lcp1^.next
                  END ;
               lcp2 := lcp1 ;
               lcp1 := lcp1^.next
             END ;
           IF sy = semicolon
             THEN
               BEGIN
                 InSymbol ;
                 Test1 ( fsys + [ident], 6 )
               END
             ELSE
               Error ( 14 )

        END ; (* WHILE *)

      IF fwptr <> nil
        THEN
          BEGIN
            Error ( 117 ) ;
            writeln ;
            writeln (listfile) ;
            REPEAT
              writeln ( ' type-id ', fwptr^.name ) ;
              EndLineList ;
              writeln ( listfile, ' type-id ', fwptr^.name ) ;
              fwptr := fwptr^.next
            UNTIL fwptr = nil ;
          END

    END ; (* TypeDeclaration *)



    PROCEDURE VarDeclaration ;
    (************************)

    (* compiles type-declaration
       Syntax-diagram checked (reserved keyword VAR already parsed):

           ------------------------<----------------------------
          v                                                     ^
          !   -------           ---      ------------      ---  !
   VAR ------! ident !---------( : )----! type-ident !----( ; )--->-
           !  -------        !  ---      ------------      ---
           ^                 v
           !  -----    ---   !
            -!ident!--( , )--
              -----    ---                                      *)
    VAR

      lcp, nxt : identptr;
      lsp : structptr ;
      lsize : addrrange ;

    BEGIN

      nxt := nil ;
      REPEAT
        REPEAT
          IF sy = ident
            THEN
              BEGIN
                new ( lcp ) ;
                lcp^.name := id ;
                lcp^.next := nxt ;
                lcp^.klass := vars ;
                lcp^.idtype := nil ;
                lcp^.vkind := actual ;
                lcp^.vlev := level ;
                Enterid ( lcp ) ;
                nxt := lcp ;
                InSymbol
              END
            ELSE
              Error ( 2 ) ;
          Test2 ( fsys + typedels + [comma,colon],6,[semicolon] ) ;
          test := sy <> comma ;
          IF not test
            THEN
              InSymbol ;
        UNTIL test ;
        Intest ( colon , 5 ) ;
        Typ ( fsys + [semicolon] + typedels, lsp, lsize ) ;
        WHILE nxt <> nil DO
          BEGIN
            nxt^.idtype := lsp ;
            IF lsp^.form = files
              THEN
                BEGIN
                  nxt^.vaddr := filenumber ;
                  IF filenumber > maxfiles
                    THEN
                      error (180) ;
                   filenumber := filenumber + 1
                 END
               ELSE
                 BEGIN
                   lc := lc + lsize ;
                   nxt^.vaddr := lc ;
                 END ;
            nxt := nxt^.next
          END ;
        IF sy = semicolon
          THEN
            BEGIN
              InSymbol ;
              Test1 ( fsys + [ident], 6 )
            END
          ELSE
            Error ( 14 )
      UNTIL ( sy <> ident ) and ( not ( sy in typedels ) ) ;
      IF fwptr <> nil
        THEN
          BEGIN
            Error (118) ;
            writeln ;
            writeln (listfile) ;
            REPEAT
              writeln (  'type-id ', fwptr^.name ) ;
              EndLineList ;
              writeln ( listfile, 'type-id ', fwptr^.name ) ;
              fwptr := fwptr^.next ;
            UNTIL fwptr = nil ;
          END
    END ; (* VARDECLARATION *)



    PROCEDURE ProcDeclaration ( fsy : symbol ) ;
    (******************************************)

    VAR

      oldlev : 0 .. maxlevel ;
      lsy : symbol ;
      lcp, lcp1 : identptr;
      lsp : structptr ;
      forw : boolean ;
      oldtop : disprange ;
      llc,lcm : addrrange ;
      markp : ^char ;        (* used for mark/release *)

    PROCEDURE ParameterList ( fsy  : setofsys ; var fpar : identptr) ;
    (****************************************************************)

     (* compiles parameter-list in a procedure or function
        declaration.
        All variables found are entered in the symbol-tree
        and the nxt^-field is used to link them together. *)



      VAR

        lcp, lcp1, lcp2, lcp3  : identptr;
        lsp : structptr ;
        lkind : idkind ;
        llc, len : addrrange ;
        count, offset : integer ;

      BEGIN

        lcp1 := nil ;
        Test1 ( fsy + [lparent], 7 ) ;
        IF sy <> lparent
          THEN
            fpar := nil
          ELSE
            BEGIN
              IF forw
                THEN
                  Error ( 119 ) ;
              InSymbol ;
              IF ( not ( sy in [ident, varsy ] ) )
                THEN
                  BEGIN
                    Error ( 7 ) ;
                    Skip ( fsys + [ident, rparent ] )
                  END ;
              WHILE sy in [ ident, varsy ] DO
                BEGIN
                  IF sy = varsy
                    THEN
                      BEGIN
                        lkind := formal ;
                        InSymbol
                      END
                    ELSE
                      lkind := actual ;
                  lcp2 := nil ;
                  count := 0 ;
                  (* enter variables in paramterlist in symbol-tree
                     and link var of same type with lcp.next *)
                  REPEAT
                    IF sy = ident
                      THEN
                        BEGIN
                          new ( lcp ) ;
                          lcp^.name := id ;
                          lcp^.idtype := nil ;
                          lcp^.klass := vars ;
                          lcp^.vkind := lkind ;
                          (* link var of same type *)
                          lcp^.next := lcp2 ;
                          lcp^.vlev := level ;
                          Enterid ( lcp ) ;
                          lcp2 := lcp ;
                          count := count + 1 ;
                          InSymbol
                        END ;
                    IF not ( sy in [comma,colon] + fsys )
                      THEN
                        BEGIN
                          Error ( 7 ) ;
                          Skip ( fsys + [comma, semicolon, rparent] )
                        END ;
                    test := sy <> comma ;
                    IF not test
                      THEN
                        InSymbol
                  UNTIL test ;
                  IF sy = colon
                    THEN
                      BEGIN
                        InSymbol ;
                        IF sy = ident
                          THEN
                            BEGIN
                              Searchid ( [types],lcp ) ;
                              lsp := lcp^.idtype ;
                              lcp3 := lcp2 ;
                              offset := 0 ;
                              len := ptrsize ;
                              IF ( lkind = actual ) and
                                 ( lsp^.form < arrays )
                                THEN
                                  IF lsp^.size = 1
                                    THEN
                                      offset := 0
                                    ELSE
                                      len := lsp^.size ;
                             lc :=lc + count * len ;
                              llc := lc ;
                              WHILE lcp2 <> nil DO
                                BEGIN
                                  lcp := lcp2 ;
                                  lcp2^.idtype := lsp ;
                                  lcp2^.vaddr := llc - offset ;
                                  llc := llc- len ;
                                  lcp2 := lcp2^.next
                                END ;
                              lcp^.next := lcp1 ;
                              lcp1 := lcp3 ;
                              InSymbol
                            END
                          ELSE
                            Error ( 2 ) ;
                        Test1 ( fsys + [semicolon,rparent] , 7 )
                      END
                    ELSE
                      Error ( 5 ) ;
                  IF sy = semicolon
                    THEN
                      BEGIN
                        InSymbol ;
                        IF not ( sy in [ident,varsy] )
                          THEN
                            BEGIN
                              Error ( 7 ) ;
                              Skip ( fsys + [ident, rparent ] )
                            END
                      END

                END ; (* WHILE *)
              IF sy = rparent
                THEN
                  BEGIN
                    InSymbol ;
                    Test1 ( fsy + fsys , 6 )
                  END
                ELSE
                  Error ( 4 ) ;
              lcp3 := nil ;

              (* Reverse pointers and reserve local cells
                 for copies of multiple values  *)

              WHILE lcp1 <> nil DO
                BEGIN
                  lcp2 := lcp1^.next ;
                  lcp1^.next := lcp3 ;
                  IF lcp1^.klass = vars
                    THEN
                      IF lcp1^.idtype <> nil
                        THEN
                          IF ( lcp1^.vkind = actual ) and
                             ( lcp1^.idtype^.form > power )
                            THEN
                              BEGIN
                                lc := lc + lcp1^.idtype^.size ;
                                lcp1^.vaddr := lc
                              END ;
                  lcp3 := lcp1 ;
                  lcp1 := lcp2

                END ; (* WHILE *)

              fpar := lcp3

            END  ;

      END ; (* ParameterList *)


    BEGIN (* ProcDeclaration *)
    (*************************)

      (* Procedure to compile a procedure or a function
         declaration.
         Note that symbols procsy or funcsy already parsed.
         Entered with current symbol is proc/func identifier.
         Syntax diagram parsed :

    --------    -------    -    ----------     -
->-!funct-id!--!parlist!--(:)--!type-ident!---(;)------
    --------    -------    -    ----------     -  ^    !
                                                  !    v
    --------    -------    -                      !    !
->-!proc-id !--!parlist!--(;)---------------------     !
    --------    -------    -                           !
                                                       !
   -----------------------<----------------------------
  !
  v
  !        -------             -
   --->---! block !-----------(;)------>----------->-
  !        -------             -           !
  v                                        ^
  !        ---------                       !
   --->---( forward )------------------>---
  !        ---------                       !
  v                                        ^
  !        --------    -    --------       !
   --->---( extern )--(=)--! number !-->---
           --------    -    --------
                                                         *)
      llc := lc ;
      lc := lcaftermarkstack ;
      IF sy <> ident
        THEN
          (* identifier of proc/func expected *)
          Error ( 2 )
        ELSE
          BEGIN

            (* Decide IF forward by searching for proc/func
               identifier in symboltable at current level.
               If not found there it can not be forward *)

            SearchSection ( display [ top ] , lcp ) ;
            IF lcp <> nil
              THEN
                BEGIN
                  IF lcp^.klass = proc
                    THEN
                      forw := ( lcp^.forwdecl ) and ( fsy = procsy )
                  ELSE IF lcp^.klass = func
                    THEN
                      forw := ( lcp^.forwdecl ) and ( fsy = funcsy )
                  ELSE
                    forw := false ;
                  IF not forw
                    THEN
                      Error ( 160 )
                      (* previous declaration was not forward *)
                END
              ELSE
                forw := false ;
            IF not forw
              THEN
                (* new procedure or function description *)
                BEGIN
                  new ( lcp ) ;
                  lcp^.name := id ;
                  lcp^.idtype := nil ;
                  lcp^.pflev := level ;
                  lcp^.pfname := nproc ;
                  nproc := nproc + 1 ;
                  lcp^.externl := false ;
                  lcp^.pfdeckind := declared ;
                  IF fsy = procsy
                    THEN
                      lcp^.klass := proc
                    ELSE
                      lcp^.klass := func ;
                  (* enter description at current level *)
                  Enterid ( lcp )
                END
              ELSE
                BEGIN

                  (* description of proc/func already present
                     including fieldlist as linked list pointed
                     at by next^ in forward declaration     *)

                  (* walk through linked list for space to reserve *)

                  lcp1 := lcp^.next ;
                  WHILE lcp1 <> nil DO
                    BEGIN
                      IF lcp1^.klass = vars
                        THEN
                          IF lcp1^.idtype <> nil
                            THEN
                              BEGIN
                                lcm := lcp1^.vaddr ;
                                IF lcm > lc
                                  THEN
                                    lc := lcm
                              END ;
                          (* next field in linked list *)
                          lcp1 := lcp1^.next
                    END ;
                END ;
            InSymbol
          END ; (* symbol is ident *)

      (* remember current level *)

      oldlev := level ;
      oldtop := top ;
      IF level < maxlevel
        THEN
          level := level + 1
        ELSE
          Error ( 173 ) ;

      (* remember current display *)

      IF top < displimit
        THEN
          BEGIN
            top := top + 1 ;
            savetop := top ;
            IF forw
              THEN
                (* already fieldlist present in tree *)
                display [ top ] := lcp^.next
              ELSE
                (* create empty tree *)
                display [ top ] := nil
          END
        ELSE
          Error ( 172 ) ;
      IF fsy = procsy
        THEN
          BEGIN
            ParameterList ( [semicolon], lcp1 ) ;
            IF not forw
              THEN
                (* link linked list to proc description *)
                lcp^.next := lcp1
          END
        ELSE
          (* function *)
          BEGIN
            ParameterList ( [semicolon,colon],lcp1 ) ;
            IF not forw
              THEN
                (* link linked list to func description *)
                lcp^.next := lcp1 ;
            IF sy <> colon
              THEN
                BEGIN
                  IF not forw
                    THEN
                      (* colon expected with function-type*)
                      Error ( 123 )
                END
              ELSE
                BEGIN
                  (* determine type of function-result *)
                  InSymbol ;
                  IF sy <> ident
                    THEN
                      BEGIN
                        (* identifier expected *)
                        Error ( 2 ) ;
                        Skip ( fsys + [semicolon] )
                      END
                    ELSE
                      BEGIN
                        IF forw
                          THEN
                            (* function-type again declared *)
                            Error ( 122 ) ;
                        Searchid ( [types],lcp1 ) ;
                        lsp := lcp1^.idtype ;
                        lcp^.idtype := lsp ;
                        IF lsp <> nil
                          THEN
                            IF lsp^.form >= power
                              THEN
                                BEGIN
                                  (* function-type must be simple *)
                                  Error ( 120 ) ;
                                  lcp^.idtype := nil
                                END ;
                        InSymbol
                      END
                END
          END ;
      Intest ( semicolon , 14 ) ;
      IF sy = forwardsy
        THEN
          BEGIN
            IF forw
              THEN
                Error ( 161 )
              ELSE
                lcp^.forwdecl := true ;
            InSymbol ;
            Intest ( semicolon, 14 ) ;
            Test1 ( fsys, 6 )
          END

      ELSE IF sy = externsy
        THEN
          BEGIN
            IF forw
              THEN Error ( 161 ) ;
            InSymbol ;
            lcp^.externl := true ;
            IF ( sy = relop ) and ( op = eqop )
              THEN
                BEGIN
                  InSymbol ;
                  IF sy = intconst
                    THEN
                      lcp^.pfname := val.ival
                    ELSE
                      Error ( 106 )
                END
              ELSE
                 Error ( 51 ) ;
            REPEAT
              InSymbol
            UNTIL sy = semicolon ;
            InSymbol ;
            Test1 ( fsys, 6 )
          END

        ELSE
          (* block *)
          BEGIN
            lcp^.forwdecl := false ;
            (* remember heap with markpointer for release *)
            new ( markp ) ;
            REPEAT
            Block ( fsys, semicolon ,lcp ) ;
            IF sy <> semicolon
              THEN
                Error ( 14 )
              ELSE
                BEGIN
                  InSymbol ;
                  IF not ( sy in [ beginsy, procsy, funcsy ] )
                    THEN
                      BEGIN
                        Error ( 6 ) ;
                        Skip ( fsys )
                      END
                END ;
            UNTIL sy in [ beginsy, procsy, funcsy ] ;
            (* release heap from markpointer *)
(*          release ( markp ) *)
          END ;

        (* restore to old level and display *)
        level := oldlev ;
        top := oldtop ;
        lc := llc


    END ; (* ProcDeclaration *)


    PROCEDURE Body ( fsys : setofsys ) ;
    (**********************************)

    VAR

      i, entname, segsize : integer ;
      lcmax, llc1 : addrrange ;
      lcp : identptr;

      PROCEDURE LDAgen  ( level : integer ; dplmt : addrrange ) ;
      (*******************************************************)

      (* generate instructions to load offset address dplmt
         for level on stack, optimize if short range *)

        BEGIN

          IF dplmt < 256
            THEN                           (* ldas *)
              BEGIN
                Bytegen ( 16  + level ) ;
                Bytegen ( dplmt )
              END
            ELSE                           (* lda  *)
              BEGIN
                Bytegen ( 32 + level ) ;
                Addrgen ( dplmt )
              END

        END ; (* LDAgen *)


      PROCEDURE LODgen ( level : integer ; var fattr : attr ) ;
      (*****************************************************)

      (* load variable value on stack of level *)

        BEGIN

          (* short range and small variable *)
          IF ( fattr.typtr^.size <= 2 ) and
             ( fattr.dplmt < 256 )
            THEN
              BEGIN
                (* LOD1, LOD2 or LOD8 byte instructions *)
                Bytegen ( 64 + 16 * fattr.typtr^.size + level ) ;
                Bytegen ( fattr.dplmt )
              END
            ELSE
              BEGIN

                (* put address on stack of data item from level *)
                LDAgen ( level, fattr.dplmt ) ;

                (* Indirectly load data on stack *)
                CASE fattr.typtr^.size OF
                  1 : Bytegen ( 154 ) ;  (* IND1 *)
                  2 : Bytegen ( 155 ) ;  (* IND2 *)
                  3 : Bytegen ( 156 )    (* IND8 *)
                END
              END

        END ; (* LODgen *)


      PROCEDURE Condgen ( fop : oprange ; var fattr : attr ) ;
      (******************************************************)

      (* generate conditional opererators,
         called with operators for 2-byte items,
         for other items calculate operator-code *)

        BEGIN

          IF fattr.typtr^.form > power
            THEN
              BEGIN
                (* arrays and records *)
                (* LEQ2 ( = $90 ) to LEQM ( = $92 ) *)
                (* LES2 ( = $93 ) to LESM ( = $95 ) *)
                (* EQU2 ( = $96 ) to EQUM ( = $98 ) *)
                Bytegen ( fop + 2 ) ;
                Addrgen ( fattr.typtr^.size )
              END
          ELSE IF fattr.typtr^.size = 1
            THEN
                (* correct to 16 bit *)
                (* LEQ2 ( = $90 ) *)
                (* LES2 ( = $93 ) *)
                (* EQU2 ( = $96 ) *)
              Bytegen ( fop )
          ELSE IF fattr.typtr^.size = 2
            THEN
                (* LEQ2 ( = $90 ) *)
                (* LES2 ( = $93 ) *)
                (* EQU2 ( = $96 ) *)
              Bytegen ( fop )
          ELSE IF fattr.typtr^.size = 8
            THEN
                (* sets, only equal operator *)
                (* EQU2 ( = $96 ) to EQUM ( = $99 ) *)
              Bytegen ( fop + 3 )
          ELSE
            Error ( 179 )           (* wrong size *)

        END ; (* Condgen *)


      PROCEDURE LDCIgen ( value : integer ) ;
      (*************************************)

        (* Generate instructions to load constant
           16 bit value on stack *)

        BEGIN

          IF ( value < 16 ) and ( value >= 0 )
            THEN
              Bytegen ( value )
          ELSE IF value < 0
            THEN
              BEGIN
                Bytegen ( 176 ) ;               (* LNC *)
                Addrgen ( -value )
              END
          ELSE
            BEGIN
              Bytegen ( 160 ) ;                 (* LDC *)
              Addrgen ( value )
            END

        END ; (* LDCIgen *)



      PROCEDURE LoadSetConstant ( setconst : intset ) ;
      (***********************************************)

      (* generate instruction to
         load constant set setconst on stack *)

        VAR

          i, k, l , n : integer ;
          b : boolean ;

        BEGIN

          n := 0 ;
          b := false ;
          for i := 0 to setmax DO
            IF i in setconst
              THEN
                n := n + 1 ;
          IF n = 0
            THEN
              Bytegen ( 193 )            (* Load empty set lns *)
          ELSE IF n < 4
            THEN
              BEGIN
                for i := 0 to setmax DO
                  IF i in setconst
                    THEN
                      BEGIN
                        LDCIgen ( i ) ;
                        Bytegen ( 174 ) ;      (* sgs *)
                        IF b
                          THEN
                            Bytegen ( 175 ) ;  (* uni *)
                        b := true
                      END
               END
          ELSE         (* n >= 4 *)
            BEGIN
              Bytegen ( 186 ) ;                (* LDCS *)
              l := 0 ;
              k := 128 ;
              for i := 0 to setmax DO
                BEGIN
                  IF k < 1
                    THEN
                      BEGIN
                        k := 128 ;
                        Bytegen ( l ) ;
                        l := 0
                      END ;
                  IF i in setconst
                    THEN
                      l := l + k ;
                  (* shift *)
                  k := k div 2
                END ;
              Bytegen ( l )
            END

        END ; (* LoadSetConstant *)



      PROCEDURE CSPgen ( stpr : integer ) ;
      (***********************************)

      (* generate instruction to call standard procedure stpr *)

        BEGIN

          Bytegen ( 189  ) ;
          Bytegen ( stpr )

        END ; (* CSPgen *)


      PROCEDURE INCgen ( inc : integer ) ;
      (**********************************)

      (* Generate instructions to increment or
         decrement integer on stack with inc *)

        BEGIN

          IF inc <> 0
            THEN
              IF inc = 1
                THEN
                  Bytegen ( 185 )                 (* inc1 *)
              ELSE IF inc = -1
                THEN
                  Bytegen ( 184 )                 (* dec1 *)
              ELSE
                IF inc > 0
                  THEN
                    BEGIN
                      Bytegen ( 180 ) ;           (* inc *)
                      Addrgen ( inc )
                    END
                  ELSE
                    BEGIN
                      Bytegen ( 179 ) ;           (* dec *)
                      Addrgen ( - inc )
                    END

        END ; (* INCgen *)


      PROCEDURE Load ;
      (**************)

      (* generate instruction to load data on stack *)

      BEGIN

      IF gattr.typtr <>  nil
        THEN
          BEGIN
            (* constant *)
            IF gattr.kind = cst
              THEN
                BEGIN
                  IF ( gattr.typtr^.form = scalar )
                    THEN
                      (* load constant on stack *)
                      LDCIgen ( gattr.cval.ival )
                  ELSE IF gattr.typtr = nilptr
                    THEN
                      (* load zero on stack *)
                     Bytegen ( 0 )
                  ELSE
                    (* load constant set on stack *)
                    LoadSetConstant ( gattr.cval.valp^.pval )
              END
            ELSE IF gattr.kind = varbl
              THEN
                IF gattr.access = drct
                  THEN
                    LODgen ( level - gattr.level,gattr )
                  ELSE
                    BEGIN
                      (* put address of data item on stack *)
                      INCgen ( gattr.dplmt ) ;
                      (* Indirectly load data item on stack *)
                      CASE gattr.typtr^.size OF
                        1 : Bytegen ( 154 ) ;    (* IND1 *)
                        2 : Bytegen ( 155 ) ;    (* IND2 *)
                        8 : Bytegen ( 156 ) ;    (* IND8 *)
                      END
                    END ;
            gattr.kind := expr
          END

      END ; (* Load *)



      PROCEDURE Store ( var fattr : attr ) ;
      (************************************)

      (* generate instruction to store data item in memory *)

      BEGIN

        IF fattr.typtr <> nil
          THEN
            IF fattr.access = drct
              THEN
                BEGIN
                  (* STR1 or LOD2 *)
                  Bytegen ( 96 + 16 * fattr.typtr^.size
                            + level - fattr.level ) ;
                  Bytegen ( fattr.dplmt )
                END
              ELSE
                IF fattr.dplmt <> 0
                  THEN
                    Error ( 178 )
                  ELSE
                    CASE fattr.typtr^.size OF
                      1 : Bytegen ( 157 ) ;          (* STO1 *)
                      2 : Bytegen ( 158 ) ;          (* STO2 *)
                      8 : Bytegen ( 159 ) ;          (* STO8 *)
                    END

      END ; (* Store *)



      PROCEDURE LoadAddress ;
      (*********************)

      (* Generate instructions to load address of variable
         or constant on stack *)

      VAR

        lvp : constptr ;
        i : integer ;

      BEGIN

        IF gattr.typtr <> nil
          THEN
            BEGIN
              IF gattr.kind = cst
                THEN
                  IF IsString ( gattr.typtr )
                    THEN
                      BEGIN
                        (* Put address of string on stack *)
                        Bytegen ( 188 ) ;         (* LCA *)
                        Bytegen ( StringSize ( gattr.cval.valp )) ;
                        lvp := gattr.cval.valp ;
                        (* Put string following opcode *)
                        REPEAT
                          for i := 1 to lvp^.slgth DO
                            Bytegen ( lvp^.sval[i]) ;
                          (* walk through linked list of string parts *)
                          lvp := lvp^.sptr
                        UNTIL lvp = nil
                      END
                    ELSE
                      Error ( 178 )
                ELSE
                  IF gattr.kind = varbl
                    THEN
                      IF gattr.access = drct
                        THEN
                          LDAgen ( level - gattr.level, gattr.dplmt )
                        ELSE
                          INCgen ( gattr.dplmt )
                    ELSE
                      Error ( 178 ) ;
              gattr.kind := varbl ;
              (* address on stack, so mark as indirect *)
              gattr.access := indrct ;
              gattr.dplmt := 0

            END

      END ; (* LoadAddress *)



      PROCEDURE FalseJumpGen  ( faddr : integer ) ;
      (*************************************)

      (* Generate instruction jump conditional
         if false value on stack jump to faddr *)

      BEGIN

        Load ;
        IF gattr.typtr <> nil
          THEN
            IF gattr.typtr <> boolptr
              THEN
                (* must have boolean result on stack *)
                Error ( 144 ) ;
        Bytegen ( 177 )   ;           (* FJP *)
        Addrgen ( faddr )

      END ; (* FalseJumpGen *)


      PROCEDURE CallUserProcGen ( fp1, fp2 : integer ) ;
      (************************************************)

      (* Generate call to user procedure *)

      BEGIN

        (* proc/func in parameter list ? *)
        IF pfcttest
          THEN
            BEGIN
              Bytegen ( fp1 ) ;
              Bytegen ( 191 )         (* CUP2 *)
            END
          ELSE
            Bytegen ( 190 ) ;         (* CUP1 *)
        Bytegen ( fp2 )

      END ; (* CallUserProcGen *)



      PROCEDURE Statement ( fsys : setofsys ) ;
      (***************************************)

      VAR lcp : identptr;

        PROCEDURE Expression ( fsys : setofsys ) ; forward ;
        (**************************************************)

        PROCEDURE Selector ( fsys : setofsys ; fcp : identptr) ;
        (******************************************************)

        (* deals with complex identifiers such as pointers,
           fields of records and indexes in arrays and
           combinations, result returned in gattr         *)


        (* structures parsed , first identifier already parsed :
                                    ---
                      -------------( ^ )----------------
                     !              ---                 !
                     ^                                  !
      ------------   !   ---     ------------     ---   v
->---! identifier !--!--( ( )---! expression !---( )  )------>
   ^  ------------   !   ---  ^  ------------  !  ---   ^  !
   !                 !        !                !        !  v
   !                 v         -------<--------         !  !
   !                 !                                  !  !
   !                 !               ---                !  !
   !                  --------------( . )---------------   !
   !                                 ---                   !
   !                                                       !
    ----------------------------------<--------------------   *)


          VAR

          lattr : attr ;
          lcp : identptr;
          lmin, lmax : integer ;

        BEGIN

          gattr.typtr := fcp^.idtype ;
          gattr.kind := varbl ;
          IF fcp^.klass = vars
            THEN
              BEGIN
                (* decide if direct or indirect access *)
                IF gattr.typtr <> nil
                  THEN
                    IF ( fcp^.vkind = actual ) and
                       ( fcp^.vaddr < 256 ) and
                       ( gattr.typtr^.size <= 2 )
                      THEN
                        BEGIN
                          gattr.access := drct ;
                          gattr.level := fcp^.vlev ;
                          gattr.dplmt := fcp^.vaddr
                        END
                      ELSE
                        BEGIN
                          gattr.access := indrct ;
                          gattr.dplmt := 0 ;
                          IF fcp^.vkind = formal
                            THEN
                              BEGIN
                                Bytegen ( 96 + level - fcp^.vlev ) ;
                                Bytegen (fcp^.vaddr )
                              END
                            ELSE
                              LDAgen ( level - fcp^.vlev, fcp^.vaddr )
                        END
              END
          ELSE IF fcp^.klass = func
            THEN
              IF fcp^.pfdeckind = standard
                THEN
                  (* assignment to standard function not allowed *)
                  Error ( 150 )
                ELSE
                  BEGIN
                    (* Implied relative address of function result *)
                    gattr.access := drct ;
                    gattr.level := fcp^.pflev + 1 ;
                    gattr.dplmt := 0
                  END ;
          (* test if not error in variable, must be '^', '(' or '.' *)
          Test1 ( selectsys + fsys, 59 ) ;
          WHILE sy in selectsys DO
            BEGIN
              IF sy = lbrack
                THEN
                  BEGIN
                    (* array element *)
                    REPEAT
                      (* all indexes separated by comma's *)
                      lattr := gattr ;
                      (* save expression result *)
                      IF lattr.typtr <> nil
                        THEN
                          IF lattr.typtr^.form <> arrays
                            THEN
                              BEGIN
                                (* type of variable is not array *)
                                Error ( 138 ) ;
                                lattr.typtr := nil
                              END ;
                      (* complex structure, remember address *)
                      LoadAddress ;
                      InSymbol ;
                      (* index is expression followed by ',' or ']' *)
                      Expression ( fsys + [comma, rbrack ] ) ;
                      IF gattr.kind <> cst
                        THEN
                          Load ;
                      IF gattr.typtr <> nil
                        THEN
                          IF gattr.typtr^.form > subrange
                            THEN
                              (* indextype must be scalar or subrange *)
                              Error ( 113 ) ;
                      IF lattr.typtr <> nil
                        THEN
                          BEGIN
                            (* check if index same type as arrayindex *)
                            IF CompTypes ( lattr.typtr^.indextype,
                                           gattr.typtr )
                              THEN
                                BEGIN
                                  IF lattr.typtr^.indextype <> nil
                                    THEN
                                      BEGIN
                                        GetBounds(
                                              lattr.typtr^.indextype,
                                                  lmin, lmax ) ;
                                        IF gattr.kind = cst
                                          THEN
                                            BEGIN
                                              (* check bounds if
                                                 constant *)
                                              IF
                                                (gattr.cval.ival < lmin)
                                                or
                                                (gattr.cval.ival > lmax)
                                                THEN
                                                  Error ( 174 ) ;
                                              gattr.cval.ival :=
                                                gattr.cval.ival - lmin
                                            END
                                          ELSE
                                            INCgen ( -lmin )
                                            (* decrement offset with
                                               lmin for actual offset *)
                                      END
                                END
                              ELSE
                                Error ( 139 ) ;
                            gattr.typtr := lattr.typtr^.stype ;
                            IF gattr.typtr <> nil
                              THEN
                                IF gattr.kind = cst
                                  THEN
                                    BEGIN
                                      gattr.cval.ival :=
                                          gattr.cval.ival *
                                             gattr.typtr^.size ;
                                      LDCIgen ( gattr.cval.ival )
                                    END
                                  ELSE
                                    BEGIN
                                      IF gattr.typtr^.size <> 1
                                        THEN
                                          BEGIN
                                            LDCIgen(gattr.typtr^.size);
                                            Bytegen ( 170 )   (* mpi *)
                                          END
                                    END ;
                            Bytegen ( 162 ) ;    (* adi *)
                            gattr.kind := varbl ;
                            gattr.access := indrct ;
                            gattr.dplmt := 0
                          END
                    UNTIL sy <> comma ;
                    (* index-list ends with ']' *)
                    Intest ( rbrack, 12 )
                  END   (* array-index *)
              ELSE IF sy = period
                THEN
                  (* record-field *)
                  BEGIN
                    IF gattr.typtr <> nil
                      THEN
                        IF gattr.typtr^.form <> records
                          THEN
                            BEGIN
                              (* not a record *)
                              Error ( 140 ) ;
                              gattr.typtr := nil
                            END ;
                    InSymbol ;
                    IF sy <> ident
                      THEN
                        Error ( 2 )
                      ELSE
                        BEGIN
                          IF gattr.typtr <> nil
                            THEN
                              BEGIN
                                (* check if known field in record *)
                                SearchSection(gattr.typtr^.fstfld,lcp) ;
                                IF lcp = nil
                                  THEN
                                    BEGIN
                                      Error ( 152 ) ;
                                      gattr.typtr := nil
                                    END
                                  ELSE
                                    BEGIN
                                      gattr.typtr := lcp^.idtype ;
                                      IF gattr.access = indrct
                                        THEN
                                          gattr.dplmt :=
                                             gattr.dplmt + lcp^.vaddr
                                         ELSE
                                           gattr.dplmt :=
                                           gattr.dplmt - lcp^.vaddr

                                     END
                              END ;
                          InSymbol
                        END
                  END   (* sy = period . *)
              ELSE
                (* must be ^ of pointer *)
                BEGIN
                  IF gattr.typtr <> nil
                    THEN
                      IF gattr.typtr^.form = pointer
                        THEN
                          BEGIN
                            Load ;
                            gattr.typtr := gattr.typtr^.stype ;
                            gattr.kind := varbl ;
                            gattr.access := indrct ;
                            gattr.dplmt := 0
                          END
                        ELSE
                          Error ( 141 ) ;
                  InSymbol
                END ;   (* sy = arrow ^ *)
              Test1 ( fsys + selectsys, 6 )

            END ; (* WHILE *)

        END ; (* Selector *)



        PROCEDURE Call ( fsys : setofsys ; fcp : identptr) ;
        (**************************************************)

        VAR

          lkey : 1 .. maxstandrd  ;

        PROCEDURE Variable ( fsys : setofsys ) ;
        (**************************************)

        (* Search if known variable at current or lower level,
           return complete description by calling selector *)

        VAR

          lcp : identptr;

        BEGIN

          IF sy = ident
            THEN
              BEGIN
                (* Look in symbol-tree *)
                Searchid ( [vars], lcp ) ;
                InSymbol
              END
            ELSE
              BEGIN
                Error ( 2 ) ;
                (* return unknown variable pointer *)
                lcp := uvarptr
              END ;
          (* get complete description *)
          Selector ( fsys, lcp )

        END ; (* Variable *)



        PROCEDURE CheckFileAddress (var fileaddr : integer) ;
        (***************************************************)

        (* Checks if first argument supplied in read or write
           is a file-identifier.
           If declared (as variable or predeclared input or output)
           then the file-number is returned else zero *)

           VAR

             fcp : identptr ;

           BEGIN

             fileaddr := 0 ;
             IF sy = ident
               THEN
                 BEGIN
                   prterr := false ;
                   Searchid ( [vars], fcp) ;
                   prterr := true ;
                   IF fcp <> nil
                     THEN
                       IF fcp^.idtype <> nil
                         THEN
                           IF fcp^.idtype^.form = files
                             THEN
                               BEGIN
                                 fileaddr := fcp^.vaddr ;
                                 Insymbol ;
                                 IF sy = comma
                                   THEN
                                     Insymbol
                               END
               END

           END ; (* CheckFileAddress *)



        PROCEDURE ReadProc ;
        (******************)

        (* Generate code for standard procedure Read and Readln

           Parsed structure, readln or read already parsed in lkey :

              ----
            -( ln )------------------->-------------------------
           ^  ----  !                                           !
           !        !            ------->----------             !
           !        !           ^                  !            !
           !        v     ---   !  ------------    v            !
 read -->----------------( ( )----( file-ident )------          !
                          ---      ------------       !         !
                                                      v         v
          -----------------<------------------------------>-----!
         v                                                      !
         !                     --------------------        ---  v
          --------------------! character-variable !------( ) )---->--
                        !  !   --------------------   ! !  ---
                        ^  v                          ^ v
                        !  !   --------------------   ! !
                        !   --! integer-variable   !--  !
                        !      --------------------     !
                        !                               !
                         --------------<----------------          *)

        VAR

          fileaddr : integer ;

        BEGIN

          IF sy <> lparent
            THEN
              BEGIN
                (* no parameters, default to input *)
                LDCIGEN ( 1 ) ;
                (* Set file address instruction *)
                BYTEGEN (194) ;
              END
            ELSE
              BEGIN
                InSymbol ;
                CheckFileAddress (fileaddr ) ;

                (* generate instruction to load filenumber on stack *)
                if fileaddr = 0
                  THEN
                    fileaddr := 1 ;  (* default to input *)
                LDCIGEN ( fileaddr ) ;
                (* Set file address instruction *)
                BYTEGEN (194) ;

                IF sy = ident
                  THEN
                    BEGIN
                      REPEAT
                        Variable ( fsys + [comma,rparent]) ;
                        (* place address of variable
                           on stack for rdc or rdi *)
                        LoadAddress ;
                        IF gattr.typtr <> nil
                          THEN
                            BEGIN
                              IF gattr.typtr^.form <= subrange
                                THEN
                                  BEGIN
                                    IF CompTypes ( intptr, gattr.typtr )
                                      THEN
                                        CSPgen ( 3 )            (* rdi *)
                                    ELSE IF CompTypes ( charptr, gattr.typtr )
                                      THEN
                                        CSPgen ( 5 )            (* rdc *)
                                    ELSE
                                      Error ( 177 )
                                  END
                                ELSE
                                  (* error in type of standard procedure *)
                                  Error ( 116 ) ;

                            END
                          ELSE
                            (* error in type of standard procedure *)
                            Error ( 116 ) ;
                       test := sy <> comma ;
                       IF not test
                         THEN
                           InSymbol
                     UNTIL test ;
                   END ;
                Intest ( rparent, 4 )
              END  ;
            IF lkey = 5
             THEN
               CSPgen ( 4 )                (* rln *)

       END ; (* ReadProc *)



       PROCEDURE WriteProc ;
       (*******************)


        (* Generate code for standard procedure Write and Writeln

        (* Parsed structure, writeln or write already parsed in lkey :
              ----
            -( ln )------------------->------------------------
           ^  ----  !                                          !
           !        !            ------->----------            !
           !        !           ^                  !           !
           !        v     ---   !  ------------    v           !
 write-->----------------( ( )----( file-ident )---------->----!
                          ---      ------------       !        !
                                                      v        v
  -------------------------<--------------------------         !
 !                                                             !
 v                                                             !
 !    -----------------                                        !
 !->-! string          !---                                    !
 !    -----------------    !   -------------->---------        !
 !                         v  ^                        !       !
 !    -----------------    !  !  ---    -------------  v  ---  v
 !->-! char-expression !->------( : )--! int-express !---( ) )---->--
 !    -----------------    !     ---    -------------     ---
 !                         ^
 !    -----------------    !
  ->-! int-expression  !---
      -----------------                                            *)

       VAR

         lsp : structptr ;
         default : boolean ;
         llkey : 1 .. maxstandrd ;
         len : addrrange ;
         fileaddr : integer ;

        BEGIN

          llkey := lkey ;
          IF sy <> lparent
            THEN
              BEGIN
                (* no parameters, default to output *)
                LDCIGEN ( 2 ) ;
                (* Set file address instruction *)
                BYTEGEN (194) ;
              END
            ELSE
              BEGIN
                InSymbol ;
                CheckFileAddress (fileaddr ) ;

                (* generate instruction to load filenumber on stack *)
                if fileaddr = 0
                  THEN
                    fileaddr := 2 ;
                LDCIGEN ( fileaddr ) ;
                (* Set file address instruction *)
                BYTEGEN (194) ;

                IF sy in facbegsys
                 THEN
                   BEGIN
                     REPEAT
                       Expression ( fsys + [ comma, colon, rparent ] ) ;
                       lsp := gattr.typtr ;
                       IF lsp <> nil
                         THEN
                           (* put variable or address of
                              variable on stack *)
                           IF lsp^.form <= subrange
                             THEN
                               Load
                             ELSE
                               LoadAddress ;
                       (* if colon then no default, load
                          expression as nr of characters *)
                       IF sy = colon
                         THEN
                           BEGIN
                             InSymbol ;
                             Expression ( fsys + [comma, rparent] ) ;
                             IF gattr.typtr <> nil
                               THEN
                                 BEGIN
                                   IF not Comptypes (gattr.typtr,intptr)
                                     THEN
                                       Error ( 116 )
                                 END
                               ELSE
                                 Error (116) ;
                             Load ;
                             default := false
                           END
                         ELSE
                           default := true ;
                       IF Comptypes (lsp, intptr)
                         THEN
                           BEGIN
                             IF default
                               THEN
                                 (* default integer in 6 char field *)
                                 LDCIgen ( 6 ) ;
                             CSPgen ( 0 )          (* wri *)
                           END
                       ELSE IF Comptypes (lsp, charptr)
                         THEN
                           BEGIN
                             IF default
                               THEN
                                 (* default character in 1 char field *)
                                 LDCIgen ( 1 ) ;
                             CSPgen ( 1 )          (* wrc *)
                           END
                       ELSE IF lsp <> nil
                         THEN
                           BEGIN
                             IF lsp^.form = scalar
                               THEN
                                 Error ( 177 )
                             ELSE IF IsString ( lsp )
                               THEN
                                 BEGIN
                                   len := lsp^.size ;
                                   IF default
                                     THEN
                                       (* default length of string *)
                                       LDCIgen ( len ) ;
                                   LDCIgen ( len ) ;
                                   CSPgen ( 2 )     (* wrs *)
                                 END
                             ELSE
                               Error ( 116 )
                           END ;
                       test := sy <> comma ;
                       IF not test
                         THEN
                           InSymbol
                     UNTIL test ;
                   END ;
                 Intest ( rparent, 4 )
              END ;
         IF llkey = 6
           THEN
             CSPgen ( 6 )                      (* wln *)

       END ; (* WriteProc *)



       PROCEDURE Newstatement ;
       (**********************)

       (* generate code for stand procedure New,
          argument of new is pointer *)

       VAR

         lsize : addrrange ;

       BEGIN

         Variable ( fsys + [rparent] ) ;
         (* pointer is address, put on stack *)
         LoadAddress ;
         lsize := 0 ;
         IF gattr.typtr <> nil
           THEN
             IF gattr.typtr^.form = pointer
               THEN
                 BEGIN
                   IF gattr.typtr^.stype <> nil
                     THEN
                       lsize := gattr.typtr^.stype^.size
                 END
               ELSE
                 Error ( 116 ) ;
         (* load size of space to allocate in heap *)
         LDCIgen ( lsize ) ;
         CSPgen ( 7 )                       (* new *)

       END ; (* NewStatement *)



       PROCEDURE ReleaseStatement ;
       (**************************)

       (* Generate code to release heap,
          release starting from pointer argument given to Release *)

       BEGIN

         Variable ( fsys + [ rparent ] ) ;
         IF gattr.typtr <> nil
           THEN
             IF gattr.typtr^.form = pointer
               THEN
                 BEGIN
                   Load ;
                   CSPgen ( 9 )                (* rst *)
                 END
               ELSE
                 Error ( 125 )

       END ; (* ReleaseStatement *)



       PROCEDURE ResetRewriteProc ;
       (**************************)

       (* Compiles Reset and Rewrite standard procedure,
          syntax checked:

                    ---      ------------     ---     -----------------
      rewrite-->---( ( )----! file-ident !---( , )---! string-constant !--->
                    ---      ------------     ---  !  -----------------   !
                                                   !                      ^
                                                   !  -----------------   !
                                                    -! string-var ident!--
                                                      -----------------

                                                                          *)
       VAR

         fileaddr,llkey  : integer ;

       BEGIN

         llkey := lkey ;
         CheckFileAddress (fileaddr) ;
         IF fileaddr = 0
           THEN
             error(22)
           ELSE
             BEGIN
               LDCIGEN (fileaddr) ;
               ByteGen ( 194 )           (* SFA *)
             END ;
         IF not (sy in facbegsys)
           THEN
             BEGIN
               error (4) ;
               insymbol
             END
           ELSE
             BEGIN
               Expression (fsys + [rparent]) ;
               IF Isstring(gattr.typtr)
                 THEN
                   BEGIN
                     (* address of string on stack *)
                     LoadAddress ;
                     (* length of string on stack *)
                     LDCIGEN (gattr.typtr^.size)
                   END
                 ELSE
                   error (23)
             END ;
         IF llkey = 7
           THEN
             CSPGEN (13)        (* Reset RSF *)
           ELSE
             CSPGEN (14) ;      (* Rewrite RWW *)


       END ; (* ResetRewriteProc *)



       PROCEDURE CloseProc ;
       (*******************)

       (* generate code for standard procedure Close *)

       VAR

         fileaddr : integer ;

       BEGIN

         CheckFileaddress (fileaddr) ;
         IF fileaddr = 0
           THEN
             BEGIN
               error(22) ;
               fileaddr := 1
             END ;
         LDCIGEN (fileaddr ) ;
         ByteGen (194) ;                 (* SFA *)
         CSPGEN (16)                     (* CLO *)

       END ; (* CloseProc *)

       PROCEDURE OrdFunc ;
       (*****************)

       (* generate code for standard function Ord
          result is integer *)

       BEGIN

         IF gattr.typtr <> nil
           THEN
             IF gattr.typtr^.form >= power
               THEN
                 Error ( 125 )  ;
         gattr.typtr := intptr

       END ; (* OrdFunc *)

       PROCEDURE SuccFunc ;
       (*****************)

       (* generate code for standard function Succ
          result is same type *)

       BEGIN

         IF gattr.typtr <> nil
           THEN
             BEGIN
               IF gattr.typtr^.form >= power
                 THEN
                   Error ( 125 )
                 ELSE
                   Bytegen ( 185 )                 (* inc1 *)
             END

       END ; (* SuccFunc *)

       PROCEDURE PredFunc ;
       (*****************)

       (* generate code for standard function Pred
          result is same type *)

       BEGIN

         IF gattr.typtr <> nil
           THEN
             BEGIN
               IF gattr.typtr^.form >= power
                 THEN
                   Error ( 125 )
                 ELSE
                   Bytegen ( 184 )                 (* dec1 *)
             END

       END ; (* PredFunc *)

       PROCEDURE ChrFunc ;
       (*****************)

       (* generate code for standard function Chr
          result is character *)

       BEGIN

         IF gattr.typtr <> nil
           THEN
             IF gattr.typtr <> intptr
               THEN
                 Error ( 125 ) ;
         gattr.typtr := charptr

       END ; (* ChrFunc *)


       PROCEDURE OddFunc ;
       (*****************)

       (* generate code for standard function Odd
          result is boolean *)

       BEGIN

         IF gattr.typtr <> nil
           THEN
             IF gattr.typtr <> intptr
               THEN
                 Error ( 125 ) ;
         CSPgen ( 12 ) ;                 (* odd *)
         gattr.typtr := boolptr

       END ; (* OddFunc *)



       PROCEDURE EofEolnStatusFunc ;
       (***************************)

       (* generate code for standard function Eof, Eoln and Status,
          result is boolean *)

       VAR

         fileaddr : integer ;

       BEGIN

         (* default to input *)
         fileaddr := 1 ;
         IF sy = lparent
           THEN
             BEGIN
               Insymbol ;
               CheckFileaddress (fileaddr) ;
               IF fileaddr = 0
                 THEN
                   BEGIN
                     fileaddr := 1 ;
                     error(22)
                   END ;
               InTest ( rparent,4)
             END ;
         LDCIGEN (fileaddr ) ;
         ByteGen (194) ;                 (* SFA *)

         IF lkey = 8
           THEN
             BEGIN
               gattr.typtr := intptr ;
               CSPGEN (15)               (* STT *)
             END
           ELSE
             BEGIN
               gattr.typtr := boolptr ;
               IF lkey = 6
                 THEN
                   CSPgen ( 10 )         (* eoln *)
                 ELSE
                   CSPgen ( 8 ) ;        (* eof *)
             END

       END ; (* EofEolnStatusFunc *)



       PROCEDURE Callnonstandard ;
       (*************************)

       (* generate to call user-defined procedure or function,
          fcp is pointer to identifier of procedure/function

          Structure parsed :

                   ------------------>--------------------
                  ^                                       !
                  !   ---       ------------       ---    v
proc/func-ident -----( ( )-----! expression !-----( ) )------>
                      ---   ^   ------------   !
                            !                  v
                             --------<---------                *)

       VAR

         nxt, lcp : identptr;
         lsp : structptr ;
         locpar, llc : addrrange ;

       BEGIN

         (* locpar is space for local parameters *)
         locpar := 0 ;
         (* parameters linked as linked list to fcp with next^ *)
         nxt := fcp^.next ;
         pfcttest := false ;
         IF not fcp^.externl
           THEN
             IF fcp^.klass = proc
               THEN
                 (* procedure needs no space for result *)
                 Bytegen ( 48 + level - fcp^.pflev )       (* msto *)
               ELSE
                 BEGIN

                 (* function has result mark stack with return-bytes MSTN *)

                   Bytegen ( 64 + level - fcp^.pflev ) ;
                   Bytegen ( fcp^.idtype^.size )         (* ret value *)

                 END ;

         (* parse parameter-list *)

         IF sy = lparent
           THEN
             BEGIN

               (* remember location-counter *)
               llc := lc ;
               REPEAT
                 InSymbol ;
                 Expression ( fsys + [comma, rparent ] ) ;
                 IF gattr.typtr <> nil
                   THEN
                     BEGIN
                       IF nxt <> nil
                         THEN
                           BEGIN
                             lsp := nxt^.idtype ;
                             IF lsp <> nil
                               THEN
                                 BEGIN
                                   IF ( nxt^.vkind = actual )
                                     THEN
                                       IF lsp^.form < arrays
                                         THEN
                                           BEGIN
                                             (* reserve room *)
                                             Load ;
                                             locpar := locpar+lsp^.size
                                           END
                                         ELSE
                                           BEGIN
                                             IF (lsp^.form= arrays) and
                                                (gattr.kind = cst)
                                               THEN
                                                 LoadAddress
                                             ELSE IF (gattr.kind=expr ) or
                                                (gattr.kind=cst )
                                               THEN
                                                 BEGIN
                                                   Load ;
                                                   LODgen (0, gattr) ;
                                                   lc := lc +
                                                     gattr.typtr^.size ;
                                                   LDAgen ( 0, lc ) ;
                                                   IF lcmax < lc
                                                     THEN
                                                       lcmax := lc ;
                                                    END
                                               ELSE
                                                 LoadAddress ;
                                             locpar := locpar + ptrsize
                                           END
                                ELSE
                                  (* formal *)
                                  IF gattr.kind = varbl
                                    THEN
                                      BEGIN
                                        LoadAddress ;
                                        locpar := locpar + ptrsize
                                      END
                                    ELSE
                                      (* formal can not be variable *)
                                      Error ( 154 ) ;
                                 IF not CompTypes (lsp,gattr.typtr)
                                   THEN
                                     (* arg and par not compatible *)
                                     Error ( 142 )
                             END
                         END
                     END ;
                   IF nxt <> nil
                     THEN
                       (* more parameters expected *)
                       nxt := nxt^.next
               UNTIL sy <> comma ;
               lc := llc ;
               Intest ( rparent, 4 )
             END ; (* sy = lparent *)
         IF nxt <> nil
           THEN
             (* more parameters expected than found *)
             Error ( 126 ) ;
         IF fcp^.externl
           THEN
             BEGIN
               (* external procedure/function *)
               LDCIgen ( locpar ) ;
               Bytegen ( 187 ) ;                (* cap *)
               Addrgen ( fcp^.pfname )
             END
           ELSE
             CallUserProcGen ( locpar, fcp^.pfname ) ;
         IF fcp^.idtype <> nil
           THEN
             IF fcp^.idtype^.size = 1
               THEN
                 (* result fixed to 2 bytes *)
                 Bytegen ( 192 ) ;              (* fix21 *)
         pfcttest := true ;
         gattr.typtr := fcp^.idtype

       END ; (* Callnonstandard *)



       BEGIN (* Call *)
       (**************)

         IF fcp^.pfdeckind <> standard
           THEN
             CallNonStandard
           ELSE
             BEGIN
               (* standard procedure and functions *)
               lkey := fcp^.key ;
               (* lkey is returned by searchid *)
               IF fcp^.klass = proc
                 THEN
                   (* standard procedures *)
                   BEGIN
                     IF lkey = 9
                       THEN
                         (* EXIT *)
                         Bytegen ( 161 )            (* retp *)
                     ELSE IF lkey = 10
                       THEN
                         (* HALT standard procedure *)
                          CSPgen ( 11 )
                     ELSE IF ( lkey = 1 ) or ( lkey = 5 )
                       THEN
                         ReadProc
                     ELSE IF ( lkey = 2 ) or ( lkey = 6 )
                       THEN
                         WriteProc
                     ELSE
                       BEGIN
                         (* arguments for procedure required *)
                         Intest ( lparent, 9 ) ;
                         IF lkey = 3
                           THEN
                             NewStatement
                         ELSE IF lkey= 4
                           THEN
                             ReleaseStatement
                         ELSE IF (lkey = 7) or (lkey = 8)
                           THEN
                             ResetRewriteProc
                         ELSE IF lkey = 11
                           THEN
                             CloseProc
                         ELSE
                           Error ( 178 ) ;
                         Intest ( rparent, 4 )
                       END
                   END
                 ELSE
                   (* standard functions *)
                   IF (lkey < 6)
                     THEN
                       BEGIN
                         (* arguments for function required *)
                         Intest ( lparent, 9 ) ;
                         Expression ( fsys + [rparent] ) ;
                         Load ;
                         CASE lkey OF
                           1 : OrdFunc  ;
                           2 : ChrFunc  ;
                           3 : OddFunc  ;
                           4 : SuccFunc ;
                           5 : PredFunc
                         END ;
                         Intest ( rparent, 4 )
                       END
                   ELSE IF (lkey = 6) OR (lkey = 7) OR (lkey = 8)
                     THEN
                       EofEolnStatusFunc

             END  (* standard procedure and functions *)

       END ; (* Call *)



       PROCEDURE Expression ;
       (********************)

         VAR

           lattr : attr ;
           lop : operator ;

         PROCEDURE SimpleExpression ( fsys : setofsys ) ;
         (**********************************************)

         var

           lattr : attr ;
           lop : operator ;
           signed : boolean ;

           PROCEDURE OPgen ( var lattr : attr ;
           (***************) opi, opb, ops : integer ) ;

           (* procedure to compile operations.
              opi, opb, ops are the operations to generate
              for operands of type integer, boolean and set
              respectively. An argument will be zero for
              operand/operator pairs which are invalid     *)

           var

             test : boolean ;

           BEGIN

             test := false ;
             IF CompTypes ( lattr.typtr, intptr ) and
                CompTypes ( gattr.typtr, intptr )
               THEN
                 BEGIN
                   IF opi <> 0
                     THEN
                       BEGIN
                         test := true ;
                         Bytegen ( opi )
                       END
                 END
             ELSE IF CompTypes ( lattr.typtr, boolptr ) and
                     CompTypes ( gattr.typtr, boolptr )
               THEN
                 BEGIN
                   IF opb <> 0
                     THEN
                       BEGIN
                         test := true ;
                         Bytegen ( opb )
                       END
                 END
             ELSE IF ( lattr.typtr^.form = power ) and
                     CompTypes ( lattr.typtr, gattr.typtr )
               THEN
                 IF ops <> 0
                   THEN
                     BEGIN
                       test := true ;
                       Bytegen ( ops )
                     END ;
             IF not test
               THEN
                 BEGIN
                   Error ( 134 ) ;
                   gattr.typtr := nil
                 END

           END ; (* OPgen *)


           PROCEDURE Term ( fsys : setofsys ) ;
           (**********************************)

           var

             lattr : attr ;
             lop : operator ;

             PROCEDURE Factor ( fsys : setofsys ) ;
             (************************************)

             var

               lcp : identptr;
               lvp : constptr ;
               varpart : boolean ;
               cstpart : intset ;
               lsp : structptr ;
               i   : integer ;

               PROCEDURE SetExpression ;
               (***********************)

               (* structure parsed :

             ---                                  ---
      --->--( [ )--------------------------------( ] )----->
             ---  !                            !  ---
                  v        ------------        ^
                   ---->--! expression !--------
                     ^     ------------    !
                     !          ---        v
                      ---------( , )-------
                                ---                     *)

               BEGIN

                 InSymbol ;
                 (* lege set *)
                 cstpart := [] ;
                 varpart := false ;
                 new ( lsp ) ;
                 lsp^.stype := nil ;
                 lsp^.size := setsize ;
                 lsp^.form := power ;
                 IF sy = rbrack
                   THEN
                     BEGIN
                       (* empty set *)
                       gattr.typtr := lsp ;
                       gattr.kind := cst ;
                       InSymbol
                     END
                   ELSE
                     BEGIN
                       REPEAT
                         Expression ( fsys + [comma,rbrack ] ) ;
                         IF gattr.typtr <> nil
                           THEN
                             IF gattr.typtr^.form > subrange
                               THEN
                                 BEGIN
                                   Error ( 136 ) ;
                                   gattr.typtr := nil
                                 END
                               ELSE
                                 IF CompTypes
                                    ( lsp^.stype, gattr.typtr )
                                   THEN
                                     BEGIN
                                       IF gattr.kind = cst
                                         THEN
                                           BEGIN
                                             i := gattr.cval.ival ;
                                             IF ( i < 0 ) OR
                                               ( i > setmax )
                                               THEN
                                                 error ( 169)
                                               ELSE
                                                 cstpart := cstpart + [i]
                                           END
                                         ELSE
                                           BEGIN
                                             Load ;
                                             Bytegen ( 174 ) ; (* sgs *)
                                             IF varpart
                                               THEN            (* uni *)
                                                 Bytegen ( 175 )
                                               ELSE
                                                 varpart := true
                                           END ;
                                       lsp^.stype := gattr.typtr ;
                                       gattr.typtr := lsp ;
                                     END
                                   ELSE
                                     Error ( 137 ) ;
                         test := sy <> comma ;
                         IF not test
                           THEN
                             InSymbol
                       UNTIL test ;
                       Intest ( rbrack, 12 )
                     END ;
                 IF varpart
                   THEN
                     BEGIN
                       IF cstpart <> []
                         THEN
                           BEGIN
                             LoadSetConstant ( cstpart ) ;
                             Bytegen ( 175 ) ;         (* uni *)
                             gattr.kind := expr
                           END
                     END
                   ELSE
                     BEGIN
                       new ( lvp ) ;
                       lvp^.pval := cstpart ;
                       gattr.cval.valp := lvp
                     END


               END ; (* SetExpression *)

               BEGIN (* Factor *)

               (* structure parsed :



                ---------------------
->-------------! constant identifier !--------------------->
   !            ---------------------                    !
   !                                                     ^
   !            ---------------------                    !
   !-----------! variable identifier !-------------------!
   !            ---------------------                    !
   !                                                     ^
   !            ---------------------       ----------   !
   !-----------! function identifier !-->--! par-list !--!
   !            ---------------------       ----------   !
   !                                                     ^
   !            ---------------------                    !
   !-----------! integer  constant   !-------------------!
   !            ---------------------                    !
   !                                                     ^
   !            ---------------------                    !
   !-----------! string constant     !-------------------!
   !            ---------------------                    !
   !                                                     ^
   !     ---    ---------------------       ---          !
   !----( ( )--! expression          !-----( ) )---------!
   !     ---    ---------------------       ---          !
   !                                                     ^
   !   -----    ---------------------                    !
   !--( not )--! factor              !-------------------!
   !   -----    ---------------------                    !
   !                                                     ^
   !    ---     ---------------------                    !
   !---( [ )---! set expression      !-------------------
        ---     ---------------------                       *)


                 IF not ( sy in facbegsys )
                   THEN
                     BEGIN
                       Error ( 58 ) ;
                       Skip ( fsys + facbegsys ) ;
                       gattr.typtr := nil
                     END ;
                 WHILE sy in facbegsys DO
                   BEGIN
                     CASE sy OF
                       ident :
                         BEGIN
                           Searchid ( [konst, vars, func], lcp ) ;
                           InSymbol ;
                           IF lcp^.klass = func
                             THEN
                               BEGIN
                                 Call ( fsys, lcp ) ;
                                 gattr.kind := expr
                               END
                           ELSE IF lcp^.klass = konst
                             THEN
                               BEGIN
                                 gattr.typtr := lcp^.idtype ;
                                 gattr.kind := cst ;
                                 gattr.cval := lcp^.values
                               END
                           ELSE
                             Selector ( fsys, lcp )
                         END ;
                       intconst :
                         BEGIN
                           gattr.typtr := intptr ;
                           gattr.kind := cst ;
                           gattr.cval := val ;
                           InSymbol
                         END ;
                       stringconst :
                         BEGIN
                           IF lgth = 1
                             THEN
                               gattr.typtr := charptr
                             ELSE
                               BEGIN
                                 new ( lsp ) ;
                                 lsp^.stype := charptr ;
                                 lsp^.form := arrays ;
                                 lsp^.indextype := nil ;
                                 lsp^.size := StringSize ( val.valp ) ;
                                 gattr.typtr := lsp
                               END ;
                           gattr.kind := cst ;
                           gattr.cval := val ;
                           InSymbol
                         END ;
                       lparent :
                         BEGIN
                           InSymbol ;
                           Expression ( fsys + [rparent ] ) ;
                           Intest ( rparent , 4 )
                         END ;
                       notsy :
                         BEGIN
                           InSymbol ;
                           Factor ( fsys ) ;
                           Load ;
                           Bytegen ( 172 ) ;      (* not *)
                           IF gattr.typtr <> nil
                             THEN
                               IF not CompTypes ( gattr.typtr, boolptr )
                                 THEN
                                   BEGIN
                                     (* result must be
                                        boolean to be negated *)
                                     Error ( 135 ) ;
                                     gattr.typtr := nil
                                   END ;
                         END ;
                       lbrack : SetExpression
                     END ;
                     Test2 ( fsys, 6, facbegsys )

                   END ; (* WHILE *)

                 END ; (* Factor *)

           BEGIN (* Term *)

           (* structure parsed :

         --------
 ---->--! factor !-------------------------------------------------->
         --------  ^               !     !      !       !       !
                   !               v     v      v       v       v
                   !              ---   ---   -----   -----   -----
                   !             ( * ) ( / ) ( div ) ( mod ) ( and )
                   !              ---   ---   -----   -----   -----
                   !   --------    !     !      !       !       !
                    <-! factor !-<----<-----<-------<------<----
                       --------                                     *)

             Factor ( fsys + [mulop] ) ;
             WHILE sy = mulop DO
               BEGIN
                 Load ;
                 lattr := gattr ;
                 lop := op ;
                 InSymbol ;
                 Factor ( fsys + [mulop] ) ;
                 Load ;
                 IF ( lattr.typtr <> nil ) and
                    ( gattr.typtr <> nil )
                   THEN
                     BEGIN
                       IF lop = mul
                         THEN
                           OPgen ( lattr, 170, 0 , 167 ) (* mpi, int *)
                       ELSE IF lop = idiv
                         THEN
                           OPgen ( lattr, 165, 0 , 0   ) (* dvi *)
                       ELSE IF lop = imod
                         THEN
                           OPgen ( lattr, 169, 0 , 0   ) (* mod *)
                       ELSE IF lop = andop
                         THEN
                           OPgen ( lattr, 0, 163,  0   ) (* and *)
                     END
                   ELSE
                     gattr.typtr := nil

               END ; (* WHILE *)

           END ; (* Term *)


         BEGIN (* SimpleExpression *)

         (* structure parsed :

            ---
        ---( + )->-
       !    ---    !
       !           v     ------
---->--!------------->--! term !---------------------------------->
       !           ^     ------  ^              !      !        !
       !    ---    !             !              v      v        v
        ---( - )->-              !             ---    ---      ----
            ---                  !            ( + )  ( - )    ( or )
                                 !             ---    ---      ----
                                 !   ------     !      !        !
                                  --! term !--------------------
                                     ------                       *)

           signed := false ;
           IF ( sy = addop ) and
              ( op in [plus, minus ] )
             THEN
               BEGIN
                 signed := ( op = minus ) ;
                 InSymbol
               END ;
           Term ( fsys + [addop] ) ;
           IF signed
             THEN
               BEGIN
                 Load ;
                 IF CompTypes ( gattr.typtr, intptr )
                   THEN
                     Bytegen ( 171 )                 (* ngi *)
                   ELSE
                     BEGIN
                       Error ( 134 ) ;
                       gattr.typtr := nil
                     END
               END ;
           WHILE sy = addop DO
             BEGIN
               (* save on stack *)
               Load ;
               (* remember previous expression result *)
               lattr := gattr ;
               lop := op ;
               InSymbol ;
               Term ( fsys + [addop ] ) ;
               Load ;
               IF ( lattr.typtr <> nil ) and
                  ( gattr.typtr <> nil )
                 THEN
                   BEGIN
                     IF lop = plus
                       THEN
                         OPgen ( lattr, 162, 0, 175 )  (* adi, uni *)
                     ELSE IF lop = minus
                       THEN
                         OPgen ( lattr, 173, 0, 164 )  (* sbi, dIF *)
                     ELSE IF lop = orop
                       THEN
                         OPgen ( lattr, 0, 168 , 0  )  (* or *)
                   END
                 ELSE
                   gattr.typtr := nil ;
               gattr.kind := expr

             END ; (* WHILE sy = addop *)

         END ; (* SimpleExpression *)


       BEGIN (* Expression *)

       (* structure parsed :

  -------------
-! simple expr !------------------------------------------------------->
  -------------  !                                                    !
                 v                                                    !
               --------------------------------------                 !
              !     !     !     !      !      !      !                !
              v     v     v     v      v      v      v                !
             ---   ---   ---   ----   ----   ----   ----              !
            ( = ) ( < ) ( > ) ( <> ) ( >= ) ( <= ) ( in )             ^
             ---   ---   ---   ----   ----   ----   ----              !
              !     !     !     !      !      !      !                !
              v     v     v     v      v      v      v  ------------  !
               ----------------------------------------! simpl expr !-
                                                        ------------  *)

         SimpleExpression ( fsys + [ relop ] ) ;
         IF sy = relop
           THEN
             BEGIN
               IF gattr.typtr <> nil
                 THEN
                   IF gattr.typtr^.form <= power
                     THEN
                       Load
                     ELSE
                       LoadAddress ;
               lattr := gattr ;
               lop := op ;
               InSymbol ;
               SimpleExpression ( fsys ) ;
               IF gattr.typtr <> nil
                 THEN
                   IF gattr.typtr^.form <= power
                     THEN
                       Load
                     ELSE
                       LoadAddress ;
               IF ( lattr.typtr <> nil ) and
                  ( gattr.typtr <> nil )
                 THEN
                   IF lop = inop
                     THEN
                       IF gattr.typtr^.form = power
                         THEN
                           IF CompTypes ( lattr.typtr,
                                          gattr.typtr^.stype )
                             THEN
                               Bytegen ( 166 )              (* inn *)
                             ELSE
                               BEGIN
                                 Error ( 129 ) ;
                                 gattr.typtr := nil
                               END
                         ELSE
                           BEGIN
                             Error ( 130 ) ;
                             gattr.typtr := nil
                           END
                     ELSE
                       BEGIN
                         IF CompTypes ( lattr.typtr, gattr.typtr )
                           THEN
                             BEGIN
                               IF lattr.typtr^.form = pointer
                                 THEN
                                   BEGIN
                                     (* test on equality allowed only *)
                                     IF lop in [ ltop,leop, gtop,geop]
                                       THEN
                                         Error ( 131 )
                                   END
                               ELSE IF lattr.typtr^.form = power
                                 THEN
                                   BEGIN
                                     (* inclusion not allowed in set comp *)
                                     IF lop in [ ltop, gtop ]
                                       THEN
                                         Error ( 132 )
                                   END
                               ELSE IF lattr.typtr^.form = arrays
                                 THEN
                                   BEGIN
                                     (* test on equality allowed only *)
                                     IF not IsString (lattr.typtr ) and
                                       (lop in [ ltop,leop, gtop,geop])
                                       THEN
                                         Error ( 131 )
                                   END
                               ELSE IF lattr.typtr^.form = records
                                 THEN
                                   BEGIN
                                     (* test on equality allowed only *)
                                     IF lop in [ ltop,leop, gtop,geop]
                                       THEN
                                         Error ( 131 )
                                   END ;
                               IF (lattr.typtr^.form = power) and
                                  (lop = leop)
                                 THEN
                                   CondGen ( 145 , lattr)  (* leq8 *)
                               ELSE IF (lattr.typtr^.form = power) and
                                  (lop = geop)
                                 THEN
                                   CondGen ( 148 , lattr)  (* geq8 *)
                               ELSE IF lop = ltop
                                 THEN
                                   Condgen ( 147, lattr )  (* les *)
                               ELSE IF lop = leop
                                 THEN
                                   Condgen ( 144, lattr )  (* leq *)
                               ELSE IF lop = eqop
                                 THEN
                                   Condgen ( 150, lattr )  (* equ *)
                               ELSE
                                 BEGIN
                                   IF lop = gtop
                                     THEN
                                       Condgen (144,lattr) (* leq *)
                                   ELSE IF lop = geop
                                     THEN
                                       Condgen (147,lattr) (* les *)
                                   ELSE IF lop = neop
                                     THEN
                                       Condgen (150,lattr) (* equ *) ;
                                   Bytegen ( 172 )         (* not *)
                                 END
                             END
                           ELSE
                             Error ( 129 )

                       END ; (* lop = inop *)
                   gattr.typtr := boolptr ;
                   gattr.kind := expr

             END  (* sy = relop *)

        END ; (* Expression *)



        PROCEDURE Assignment ( fcp : identptr) ;
        (**************************************)

        (* structure parsed :

                ------------       ---      ------------
        ->-----! identifier !-----( = )----! expression !---->
                ------------       ---      ------------
                                                           *)

        VAR

          lmax  : integer ; (* dummy for GetBounds *)
          lattr : attr ;

        BEGIN

          (* determine type of identifier for assignment *)
          Selector ( fsys + [ becomes ], fcp ) ;

          IF sy <> becomes
            THEN
              (* := expected *)
              Error ( 51 )
            ELSE
              BEGIN
                IF gattr.typtr <> nil
                  THEN
                    IF ( gattr.access <> drct ) or
                       ( gattr.typtr^.form > power )
                      THEN
                        LoadAddress ;
                lattr := gattr ;
                InSymbol ;
                Expression ( fsys ) ;
                IF gattr.typtr <> nil
                  THEN
                    (* complex structure not simple store *)
                    IF gattr.typtr^.form  <= power
                      THEN
                        Load
                      ELSE
                        LoadAddress ;
                IF ( lattr.typtr <> nil ) and
                   ( gattr.typtr <> nil )
                  THEN
                    BEGIN
                      IF CompTypes ( lattr.typtr, gattr.typtr )
                        THEN
                          BEGIN
                            IF lattr.typtr^.form in [arrays, records]
                              THEN
                                BEGIN
                                  Bytegen ( 183 ) ;            (* mov *)
                                  Addrgen ( lattr.typtr^.size )
                                END
                              ELSE
                                Store ( lattr )
                          END
                        (* assignment of array of char to array of char *)
                      ELSE IF (isstring(lattr.typtr))
                          and (isstring(gattr.typtr))
                        THEN
			  BEGIN
                            IF lattr.typtr^.size < gattr.typtr^.size
                              THEN
                                BEGIN
                                  Bytegen ( 183 ) ;            (* mov *)
                                  Addrgen ( lattr.typtr^.size )
                                END
                              ELSE
                                BEGIN
                                  (* less to move, add blanks to end *)
                                  Bytegen ( 196 ) ;           (* mvb *)
                                  Addrgen ( gattr.typtr^.size ) ;
                                  Addrgen ( lattr.typtr^.size -
                                            gattr.typtr^.size )
                                END
                          END
                        (* assignment of char to array of char *)
                      ELSE IF (isstring(lattr.typtr))
                          and (gattr.typtr = charptr)
                        THEN
			  BEGIN
                            Bytegen ( 196 ) ;           (* mvb *)
                            Addrgen ( 1 ) ;
                            Addrgen ( lattr.typtr^.size - 1)
                          END
                        ELSE
                          Error ( 129 )
                    END
              END (* sy = becomes *)

        END ; (* Assignment *)



        PROCEDURE CompoundStatement ;
        (***************************)

        (* structure parsed :

             -------       -----------        -----
   ---->--- ( begin )-----! statement !------( end )---->
             -------   ^   -----------   !    -----
                       !                 v
                       !       ---       !
                        ------( ; )------
                               ---                       *)
        BEGIN

          REPEAT
            REPEAT
              Statement ( fsys + [semicolon, endsy ] )
            UNTIL not ( sy in statbegsys ) ;
            test := sy <> semicolon ;
            IF not test
              THEN
                InSymbol
          UNTIL test ;
          Intest ( endsy, 13 )

        END ; (* CompoundStatement *)



        PROCEDURE IfStatement ;
        (*********************)


        (* Parsed by IFStatement :
           ( IF symbol already parsed)
                                               ----   ---------
                                            --(ELSE)-!statement!-
                                           !   ----   ---------  !
         ----------    ----    ---------   !                     v
 if ->--!expression!--(THEN)--!statement!--!                     -->
         ----------    ----    ---------   !                     ^
                                           !                     !
                                            ---------------------

                                                *)

        VAR

          icix1, icix2 : integer ;

        BEGIN

          (* expression must give boolean result for cond. jump *)
          Expression ( fsys + [thensy] ) ;
          FalseJumpGen ( 0 ) ;
          (* remember address for jump to plant after statement *)
          icix1 := ic + icn - 2 ;

          (* Then symbol *)
          Intest ( thensy, 52 ) ;

          (* statement *)
          Statement ( fsys + [elsesy] ) ;

          (* maybe else part with statement *)
          IF sy = elsesy
            THEN
              BEGIN
                (* end then part with unconditional jump to end *)
                icix2 := ic + icn ;
                GenUJPent ( 178 , 0 ) ;     (* ujp *)
                (* fill in address of cond jump if result is false *)
                PlantAddress (  icix1, ic + icn ) ;
                InSymbol ;
                Statement ( fsys ) ;
                (* we know now end of else part *)
                PlantAddress ( icix2 + 1, ic + icn )
              END
            ELSE
              (* no else part, fill in conditional jump address *)
              PlantAddress ( icix1, ic + icn )

        END ; (* IfStatement *)



        PROCEDURE CaseStatement ;
        (***********************)

        (* Parsed by CaseStatement :
           ( Case symbol already parsed)

            -----------     ----
case->----! expression !---( OF ) ---->---
           ------------     ----          !
                                          !
    --------------------------------------
   !
   v
   !
    -----------! constant !------( ; ) ---! statement !---------
     ^     ^    ----------   !    ---      -----------       !  !
     !     !                 v                               v  v
     !     !        ---      !                               !  !
     !      -------( , )-----                                !  !
     !              ---              ---                     !  !
      ------------------------------( ; )--------------------   !
                                     ---                        !
    ------------------------------------------------------------
   !
   v                ------
   !          -----( ELSE )----
   !         !      ------     !
   !         !                 !     -----------
    ---->----!                 !-->-! statement ! -->-------->
   !         !                 !     -----------         !
   !         !   -----------   !                         !
   !          --( OTHERWISE )--                          ^
   !             -----------                             !
   !                                                     !
    ---->------------------------------------------------  *)

        TYPE

          CasInfoptr = ^caseinfo ;
          caseinfo    = PACKED RECORD
                          casenext : CasInfoptr ;
                          csstart,cslab : integer
                        END ;

        var

          lsp, lsp1 : structptr ;
          fstptr, lpt1, lpt2, lpt3 : CasInfoptr  ;
          lval : valu ;
          lcix, lcix1, lmin, lmax : integer ;

        BEGIN

          Expression ( fsys + [ofsy, comma, colon ] ) ;
          Load ;
          lcix := ic + icn ;
          GenUJPent ( 178, 0 ) ;             (* ujp *)
          GenUJPent ( 178, 0 ) ;             (* ujp *)
          lsp := gattr.typtr ;
          IF lsp <> nil
            THEN
              IF lsp^.form > subrange
                THEN
                  BEGIN
                    Error ( 144 ) ;
                    lsp := nil
                  END ;
          Intest ( ofsy, 8 ) ;
          fstptr := nil ;
          REPEAT
            lpt3 := nil ;
            lcix1 := ic + icn ;
            REPEAT
              Constant ( fsys + [comma,colon], lsp1, lval ) ;
              IF lsp1 <> nil
                THEN
                  IF CompTypes ( lsp, lsp1 )
                    THEN
                      BEGIN
                        lpt1 := fstptr ;
                        lpt2 := nil ;
                        test := true ;
                        WHILE ( lpt1 <> nil ) and test DO
                          IF lpt1^.cslab <= lval.ival
                            THEN
                              BEGIN
                                IF lpt1^.cslab = lval.ival
                                  THEN
                                    Error ( 156 ) ;
                                test := false
                              END
                            ELSE
                              BEGIN
                                lpt2 := lpt1 ;
                                lpt1 := lpt1^.casenext
                              END ;
                        new ( lpt3 ) ;
                        lpt3^.casenext := lpt1 ;
                        lpt3^.cslab := lval.ival ;
                        lpt3^.csstart := lcix1 ;
                        IF lpt2 = nil
                          THEN
                            fstptr := lpt3
                          ELSE
                            lpt2^.casenext := lpt3
                      END
                    ELSE
                      Error ( 147 ) ;
                test := sy <> comma ;
                IF not test
                  THEN
                    InSymbol
            UNTIL test ;
            Intest ( colon, 5 ) ;
            REPEAT
              Statement ( fsys + [semicolon, elsesy ] )
            UNTIL not ( sy in statbegsys ) ;
            IF lpt3 <> nil
              THEN
                GenUJPent ( 178, lcix + 3 ) ;    (* ujp *)
            test := sy <> semicolon ;
            IF not test
              THEN
                InSymbol ;
          UNTIL test or ( sy = endsy ) or ( sy = elsesy ) ;
          PlantAddress ( lcix + 1 , ic + icn ) ;
          IF fstptr <> nil
            THEN
              BEGIN
                lmax := fstptr^.cslab ;
                lpt1 := fstptr ;
                fstptr := nil ;
                (* reverse pointers *)
                REPEAT
                  lpt2 := lpt1^.casenext ;
                  lpt1^.casenext := fstptr ;
                  fstptr := lpt1 ;
                  lpt1 := lpt2
                UNTIL lpt1 = nil ;
                lmin := fstptr^.cslab ;
                Bytegen ( 182 )  ;          (* CAS *)
                Addrgen ( lmin ) ;
                Addrgen ( lmax ) ;
                lcix1 := ic + icn ;
                Addrgen ( 0 ) ;
                REPEAT
                  WHILE fstptr ^.cslab > lmin DO
                    BEGIN
                      Addrgen ( 0 ) ;
                      lmin := lmin + 1 ;
                    END ;
                  Addrgen ( fstptr^.csstart ) ;
                  fstptr := fstptr^.casenext ;
                  lmin := lmin + 1
                UNTIL fstptr = nil
              END (* fstptr <> nil *)
            ELSE
              Error ( 157 ) ;
          IF sy = elsesy
            THEN
              BEGIN
                InSymbol ;
                PlantAddress ( lcix1, ic + icn ) ;
                REPEAT
                  Statement ( fsys + [semicolon] )
                UNTIL not ( sy in statbegsys ) ;
                GenUJPent ( 178 , lcix + 3 )
              END ;
          PlantAddress ( lcix + 4, ic + icn ) ;
          Intest ( endsy, 13 )

        END ; (* CaseStatement *)



        PROCEDURE RepeatStatement ;
        (*************************)

        (* Parsed by RepeatStatement :
           ( REPEAT symbol already parsed)

             -----------         -------       ------------
repeat->----! statement !-------( UNTIL ) ----! expression !--->
         ^   -----------     !   -------       ------------
         !         -         !
          --------(;)--------
                   -
                                                *)
          VAR

            laddr : integer ;

          BEGIN

            (* remember address for false jump at until *)
            laddr := ic + icn ;

            (* one or more statement blocks *)
            REPEAT
              REPEAT
                Statement ( fsys + [ semicolon, untilsy ] )
              UNTIL not ( sy in statbegsys ) ;
              test := sy <> semicolon ;
              IF not test
                THEN
                  InSymbol
            UNTIL test ;

            (* UNTIL part *)
            IF sy = untilsy
              THEN
                BEGIN
                  InSymbol ;
                  (* expression *)
                  Expression ( fsys ) ;
                  (* expression result must be boolean for FJP *)
                  FalseJumpGen ( laddr )
                END
              ELSE
                (* UNTIL expected but not found *)
                Error ( 53 )

        END ; (* RepeatStatement *)



        PROCEDURE WhileStatement ;
        (************************)

        (* Parsed by WhileStatement :
           ( while symbol already parsed)

              ------------           ----       ----------
 while ->----! expression !---------( DO ) ----!statement !--->
              ------------           ----       ----------

                                                *)
        VAR

          laddr, lcix : integer ;

        BEGIN

          (* remember address for jump at at end of while *)
          laddr := ic + icn ;

          (* parse boolean expression *)
          Expression ( fsys + [dosy] ) ;

          (* remember address of conditional jump out of loop *)
          lcix := ic + icn ;
          FalseJumpGen ( 0 ) ;

          (* DO symbol expected *)
          Intest ( dosy, 54 ) ;

          (* parse body of loop *)
          Statement ( fsys ) ;

          (* Unconditional jump back to begin of loop *)
          GenUJPent ( 178, laddr ) ;     (* ujp *)

          (* address for jump out of loop now known *)
          PlantAddress ( lcix + 1 , ic + icn ) ;

        END ; (* WhileStatement *)



        PROCEDURE ForStatement ;
        (**********************)

        (* Parsed by ForStatement :
           ( for symbol already parsed)

         -----------    ----     ------------
for ->--! var-ident !--( := )---! expression !--->
         -----------    ----     ------------     !
                                                  !
  ------------------------------------------------
 !        -------
 !    --( DOWNTO )---
 !   !   --------    !   ------------    ----    -----------
  ->-!               !--! expression !--( DO )--! statement !--->
     !     ----      !   ------------    ----    -----------
      --- ( TO )-----
           ----                                 *)

        VAR

          lattr : attr ;
          lsp : structptr ;
          lsy : symbol ;
          lcix, laddr : integer ;

        BEGIN

          (* Variable identifier, find description *)
          IF sy = ident
            THEN
              BEGIN
                Searchid ( [ vars ], lcp ) ;
                lattr.typtr := lcp^.idtype ;
                lattr.kind := varbl ;
                IF lcp^.vkind = actual
                  THEN
                    LDAgen ( level - lcp^.vlev, lcp^.vaddr )
                  ELSE
                    BEGIN
                      Error ( 155 ) ;
                      lattr.typtr := nil
                    END ;
                IF lattr. typtr <> nil
                  THEN
                    IF ( lattr.typtr^.form > subrange )
                      THEN
                        BEGIN
                          (* illegal type of loop control variable *)
                          Error ( 143 ) ;
                          lattr.typtr := nil
                        END ;
                InSymbol ;
              END
            ELSE
              BEGIN
                (* identifier expected *)
                Error ( 2 ) ;
                Skip ( fsys + [becomes, tosy, downtosy, dosy ] )
              END ;

          (* := *)
          IF sy = becomes
            THEN
              BEGIN
                InSymbol ;
                (* expression, must be same type as var. ident *)
                Expression ( fsys + [tosy, downtosy, dosy ] ) ;
                IF gattr.typtr <> nil
                  THEN
                    IF gattr.typtr^.form > subrange
                      THEN
                        (* illegal type of expression *)
                        Error ( 144 )
                      ELSE
                        IF CompTypes ( lattr.typtr, gattr.typtr )
                          THEN
                            (* put beginvalue of loop variable on stack *)
                            Load
                          ELSE
                            (* type conflict *)
                            Error ( 145 )
              END
            ELSE
              BEGIN
                (* := expected *)
                Error ( 51 ) ;
                Skip ( fsys + [tosy, downtosy, dosy ] )
              END ;

          (* parse TO or downto *)
          IF sy in [tosy, downtosy]
            THEN
              BEGIN
                lsy := sy ;
                InSymbol ;
                Expression ( fsys + [dosy] ) ;
                IF gattr.typtr <> nil
                  THEN
                    IF gattr.typtr^.form > subrange
                      THEN
                        Error ( 144 )
                      ELSE
                        IF CompTypes ( lattr.typtr, gattr.typtr )
                          THEN
                            BEGIN
                              (* put end value of loop variable on stack *)
                              Load ;
                              lcix := lattr.typtr^.size - 1 ;
                              IF lsy = downtosy
                                THEN
                                 lcix := lcix + 4 ;
                              (* stepsize on stack *)
                              Bytegen ( lcix ) ;
                              laddr := ic + icn ;
                              (* generate jump instruction to FOR statement *)
                              GenUJPent ( 178, 0 )      (* ujp *)
                            END
                          ELSE
                            (* type conflict with loop variable *)
                            Error ( 145 )
              END
            ELSE
              BEGIN
                (* to/downto expected *)
                Error ( 55 ) ;
                Skip ( fsys + [dosy] )
              END ;

          (* skip and test DO symbol *)
          Intest ( dosy, 54 ) ;

          (* statement *)
          Statement ( fsys ) ;

          (* now we know address of end of for loop *)
          PlantAddress ( laddr + 1 , ic + icn ) ;

          GenUJPent ( 145 , laddr + 3 )  ;             (* for *)

        END ; (* ForStatement *)


      BEGIN (* Statement *)
      (*******************)

        IF not ( sy in fsys + [ident] )
          THEN
            BEGIN
              Error ( 6 ) ;
              Skip ( fsys )
            END ;
        IF sy  in ( statbegsys + [ident] )
          THEN
            BEGIN
              pfcttest := false ;
              CASE sy OF
                ident :
                  BEGIN
                    Searchid ( [vars, func, proc ], lcp ) ;
                    InSymbol ;
                    IF lcp^.klass = proc
                      THEN
                        Call ( fsys, lcp )
                      ELSE
                        Assignment ( lcp )
                  END ;
                beginsy :
                  BEGIN
                    InSymbol ;
                    CompoundStatement
                  END ;
                ifsy :
                  BEGIN
                    InSymbol ;
                    IfStatement
                  END ;
                whilesy :
                  BEGIN
                    InSymbol ;
                    WhileStatement
                  END ;
                repeatsy :
                  BEGIN
                    InSymbol ;
                    RepeatStatement
                  END ;
                forsy :
                  BEGIN
                    InSymbol ;
                    ForStatement
                  END ;
                casesy :
                  BEGIN
                    InSymbol ;
                    CaseStatement
                  END
              END ;
              IF not ( sy in [semicolon,endsy,elsesy,untilsy] )
                THEN
                  BEGIN
                    Error ( 6 ) ;
                    Skip ( fsys )
                  END
            END ;

      END ; (* Statement *)


    BEGIN (* body *)
    (**************)

      (* fprocp is pointer to proc/func/program identifier *)
      entname := fprocp^.pfname ;
      (* empty code buffer *)
      WriteOut ;
      sumcheck := 0 ;
      write ( outfile,'P4') ;
      HexOut ( entname ) ;
      for i := 1 to 8 DO
        BEGIN
          write ( outfile, fprocp^.name[i] ) ;
          sumcheck := sumcheck + ord ( fprocp^.name[i])
        END ;
      HexOut (( 16383 - sumcheck ) mod 256 ) ;
      writeln(outfile ) ;

      (* remember where to put nr of bytes for local variables *)

      segsize := ic + icn ;
      GenUJPent ( 181 , 0 ) ;               (* ENT *)

      (* start with 6 bytes on stack *)
      llc1 := lcaftermarkstack ;
      (* parameter-list as linked list on next^ *)
      lcp := fprocp^.next ;
      WHILE lcp <> nil DO
        BEGIN
          IF lcp^.klass = vars
            THEN
              IF lcp^.idtype <> nil
                THEN
                  IF  lcp^.idtype^.form > power
                    THEN
                      BEGIN
                        llc1 := llc1 + ptrsize ;
                        IF lcp^.vkind = actual
                          THEN
                            BEGIN
                              (* copy record or array *)
                              LDAgen ( 0, lcp^.vaddr ) ;
                              Bytegen ( 96 ) ; (* LOD *)
                              Bytegen ( llc1 ) ;
                              Bytegen ( 183 ) ;     (* mov *)
                              Addrgen ( lcp^.idtype^.size )
                            END
                      END
                    ELSE
                      (* keep track of local space on stack *)
                      llc1 := llc1 + lcp^.idtype^.size ;
              lcp:= lcp^.next ;
        END ; (* WHILE *)
      lcmax := lc ;
      REPEAT
        REPEAT
          Statement ( fsys + [semicolon, endsy ] )
        UNTIL not ( sy in statbegsys ) ;
        test := sy <> semicolon ;
        IF not test
          THEN
            InSymbol
      UNTIL test ;
      Intest ( endsy, 13 ) ;
      Bytegen ( 161 ) ;                (* retp *)
      PlantAddress ( segsize + 1 , lcmax - lcaftermarkstack) ;

    END ; (* body *)

  BEGIN (* Block*)
  (**************)

    declpart := true ;

    REPEAT
      IF sy = constsy
        THEN
          BEGIN
            InSymbol ;
            ConstDeclaration
          END  ;
      IF sy = typesy
        THEN
          BEGIN
            InSymbol ;
            TypeDeclaration
          END  ;
      IF sy = varsy
        THEN
          BEGIN
            InSymbol ;
            VarDeclaration
          END  ;
      WHILE sy in [ procsy, funcsy ] DO
        BEGIN
          lsy := sy ;
          InSymbol ;
          ProcDeclaration( lsy )
        END ;
      IF sy <> beginsy
        THEN
          BEGIN
            Error(18) ;
            Skip ( fsys )
          END
    UNTIL sy in statbegsys ;

    declpart := false ;
    Intest ( beginsy,17) ;
    REPEAT
      body( fsys ) ;
      Test2 ( [fsy], 6, fsys )
    UNTIL ( sy = fsy ) or ( sy in blockbegsys ) ;

  END ; (* Block*)




  PROCEDURE StdNames ;
  (******************)

  BEGIN

    na[ 0] := 'false   ' ; na[ 1] := 'true    ' ; na[ 2] := 'read    ' ;
    na[ 3] := 'write   ' ; na[ 4] := 'new     ' ; na[ 5] := 'release ' ;
    na[ 6] := 'readln  ' ; na[ 7] := 'writeln ' ; na[ 8] := 'reset   ' ;
    na[ 9] := 'rewrite ' ; na[10] := 'exit    ' ; na[11] := 'halt    ' ;
    na[12] := 'close   ' ; na[13] := 'ord     ' ; na[14] := 'chr     ' ;
    na[15] := 'odd     ' ; na[16] := 'succ    ' ; na[17] := 'pred    ' ;
    na[18] := 'eoln    ' ; na[19] := 'eof     ' ; na[20] := 'status  ' ;

  END ; (* StdNames *)



  PROCEDURE Enterstdtypes ;
  (***********************)

  BEGIN

    (* integer *)

    new (intptr ) ;
    intptr^.size  := intsize ;
    intptr^.form  := scalar  ;
    intptr^.stype := nil     ;

    (* char *)

    new (charptr ) ;
    charptr^.size  := charsize ;
    charptr^.form  := scalar  ;
    charptr^.stype := nil     ;

    (* boolean *)

    new (boolptr ) ;
    boolptr^.size  := boolsize ;
    boolptr^.form  := scalar  ;
    boolptr^.stype := nil     ;

    (* pointer nil *)

    new (nilptr ) ;
    nilptr^.size  := ptrsize ;
    nilptr^.form  := pointer ;
    nilptr^.stype := nil     ;

    (* file *)

    new (fileptr ) ;
    fileptr^.size  := 0       ;
    fileptr^.stype := nil     ;
    fileptr^.form  := files   ;

  END ; (* Enterstdtypes *)




  PROCEDURE Enterstnames ;
  (*********************)

  var cp, cp1 : identptr;
      i : integer ;

  BEGIN

    (* integer *)

    new (cp) ;
    cp^.name := 'integer ' ;
    cp^.idtype := intptr ;
    cp^.next := nil ;
    cp^.klass := types ;
    Enterid (cp) ;

    (* char *)

    new (cp) ;
    cp^.name := 'char    ' ;
    cp^.idtype := charptr ;
    cp^.next := nil ;
    cp^.klass := types ;
    Enterid (cp) ;

    (* boolean *)

    new (cp) ;
    cp^.name := 'boolean ' ;
    cp^.idtype := boolptr ;
    cp^.next := nil ;
    cp^.klass := types ;
    Enterid (cp) ;

    (* text-file *)

    new (cp) ;
    cp^.name := 'text    ' ;
    cp^.idtype := fileptr ;
    cp^.next := nil ;
    cp^.klass := types;
    Enterid ( cp) ;

    (* false, true *)

    cp1 := nil ;
    for i := 0 to 1 DO
      BEGIN
        new (cp) ;
        cp^.name := na[ i ] ;
        cp^.idtype := boolptr ;
        cp^.next := cp1 ;
        cp^.klass := konst ;
          cp^.values.ival := i ;
        Enterid ( cp ) ;
        cp1 := cp
      END ;

    (* nil pointer *)

    new (cp) ;
    cp^.name := 'nil     ' ;
    cp^.idtype := nilptr ;
    cp^.next := nil ;
    cp^.klass := konst ;
      cp^.values.ival := 0 ;
    Enterid ( cp) ;


    (* i/o and pointer procedures
       key  1 = read
       key  2 = write
       key  3 = new
       key  4 = release
       key  5 = readln
       key  6 = writeln
       key  7 = reset
       key  8 = rewrite
       key  9 = exit
       key 10 = halt
       key 11 = close
*)


    for i := 2 to 12 DO
      BEGIN
        new (cp) ;
        cp^.name := na [ i ] ;
        cp^.idtype := nil ;
        cp^.next := nil ;
        cp^.klass := proc ;
          cp^.pfdeckind := standard ;
            cp^.key := i - 1 ;
        Enterid (cp)
      END ;

    (* Standard functions :
       key 1 = ord
       key 2 = chr
       key 3 = odd
       key 4 = succ
       key 5 = pred
       key 6 = eoln
       key 7 = eof
       key 8 = status        *)

    for i := 13 to 20 DO
      BEGIN
        new (cp) ;
        cp^.name := na [ i ] ;
        cp^.idtype := nil ;
        cp^.next := nil ;
        cp^.klass := func ;
          cp^.pfdeckind := standard ;
            cp^.key := i - 12 ;
        Enterid (cp)
      END ;

    (* files input and output and keyboard *)

    new (cp) ;
    cp^.name := 'input   ' ;
    cp^.idtype := fileptr ;
    cp^.next := nil ;
    cp^.klass := vars ;
      cp^.vkind := actual ;
      cp^.vaddr := 1 ;        (* filenumber 1 = input *)
      cp^.vlev := 0 ;
    Enterid (cp) ;

    new (cp) ;
    cp^.name := 'output  ' ;
    cp^.idtype := fileptr ;
    cp^.next := nil ;
    cp^.klass := vars ;
      cp^.vkind := actual ;
      cp^.vaddr := 2 ;        (* filenumber 2 = output *)
      cp^.vlev := 0 ;
    Enterid (cp) ;

    new (cp) ;
    cp^.name := 'keyboard' ;
    cp^.idtype := fileptr ;
    cp^.next := nil ;
    cp^.klass := vars ;
      cp^.vkind := actual ;
      cp^.vaddr := 3 ;        (* filenumber 3 = keyboard *)
      cp^.vlev := 0 ;
    Enterid (cp) ;

  END ; (* Enterstnames *)



  PROCEDURE Enterundecl ;
  (*********************)

  BEGIN

    new ( utypptr ) ;
    utypptr^.name := '        ' ;
    utypptr^.idtype := nil ;
    utypptr^.klass := types ;

    new (ucsptr) ;
    ucsptr^.name := '        ' ;
    ucsptr^.idtype := nil ;
    ucsptr^.next := nil ;
    ucsptr^.klass := konst ;
      ucsptr^.values.ival := 0 ;

    new (uvarptr) ;
    uvarptr^.name := '        ' ;
    uvarptr^.idtype := nil ;
    uvarptr^.next := nil ;
    uvarptr^.klass := vars ;
      uvarptr^.vkind := actual ;
      uvarptr^.vaddr := 0 ;
      uvarptr^.vlev := 0 ;

    new (ufldptr ) ;
    ufldptr^.name := '        ' ;
    ufldptr^.idtype := nil ;
    ufldptr^.next := nil ;
    ufldptr^.klass := field ;
      ufldptr^.vaddr := 0 ;

    new ( uprcptr ) ;
    uprcptr^.name := '        ' ;
    uprcptr^.idtype := nil ;
    uprcptr^.forwdecl := false ;
    uprcptr^.next := nil ;
    uprcptr^.klass := proc ;
      uprcptr^.pfdeckind := declared ;
        uprcptr^.pflev := 0 ;
        uprcptr^.pfname := 0 ;
        uprcptr^.externl := false ;

    new ( ufctptr ) ;
    ufctptr^.name := '        ' ;
    ufctptr^.idtype := nil ;
    ufctptr^.next := nil ;
    ufctptr^.forwdecl := false ;
    ufctptr^.klass := func ;
      ufctptr^.pfdeckind := declared ;
        ufctptr^.pflev := 0 ;
        ufctptr^.pfname := 0 ;
        ufctptr^.externl := false ;

  END ; (* Enterundecl *)



PROCEDURE FillErrorMessages ;
(***************************)

  BEGIN
    cmperror [   2 ] := 'Syntax: identifier expected             ' ;
    cmperror [   3 ] := 'Syntax: Program expected                ' ;
    cmperror [   4 ] := 'Syntax: ")" expected                    ' ;
    cmperror [   5 ] := 'Syntax: ":" exepected                   ' ;
    cmperror [   6 ] := 'Syntax: illegal symbol                  ' ;
    cmperror [   7 ] := 'Syntax: actual parameter list           ' ;
    cmperror [   8 ] := 'Syntax: OF expected                     ' ;
    cmperror [   9 ] := 'Syntax: "(" expected                    ' ;
    cmperror [  10 ] := 'Syntax: type specfication expected      ' ;
    cmperror [  11 ] := 'Syntax: "[" expected                    ' ;
    cmperror [  12 ] := 'Syntax: "]" expected                    ' ;
    cmperror [  13 ] := 'Syntax: END expected                    ' ;
    cmperror [  14 ] := 'Syntax: ";" expected                    ' ;
    cmperror [  15 ] := 'Syntax: integer expected                ' ;
    cmperror [  16 ] := 'Syntax: "-" expected                    ' ;
    cmperror [  17 ] := 'Syntax: BEGIN expected                  ' ;
    cmperror [  18 ] := 'Syntax: error in declaration part       ' ;
    cmperror [  19 ] := 'Syntax: error in field list             ' ;
    cmperror [  20 ] := 'Syntax: "," expected                    ' ;
    cmperror [  21 ] := 'Syntax: "*" expected                    ' ;

    cmperror [  50 ] := 'Syntax: "error in constant              ' ;
    cmperror [  51 ] := 'Syntax: ":=" expected                   ' ;
    cmperror [  52 ] := 'Syntax: THEN expected                   ' ;
    cmperror [  53 ] := 'Syntax: UNTIL expected                  ' ;
    cmperror [  54 ] := 'Syntax: DO expected                     ' ;
    cmperror [  55 ] := 'Syntax: TO/DOWNTO expected              ' ;
    cmperror [  56 ] := 'Syntax: IF expected                     ' ;
    cmperror [  58 ] := 'Syntax: ill-formed expression           ' ;
    cmperror [  59 ] := 'Syntax: error in variable               ' ;

    cmperror [ 101 ] := 'Identifier declared twice               ' ;
    cmperror [ 102 ] := 'Low bound exceeds high-bound            ' ;
    cmperror [ 103 ] := 'Identifier is not a type identifier     ' ;
    cmperror [ 104 ] := 'Identifier not declared                 ' ;
    cmperror [ 105 ] := 'Sign not allowed                        ' ;
    cmperror [ 106 ] := 'Number expected                         ' ;
    cmperror [ 107 ] := 'Incompatible subrange types             ' ;
    cmperror [ 110 ] := 'Tag type must be an ordinal type        ' ;
    cmperror [ 111 ] := 'Incompatible with tag type              ' ;
    cmperror [ 113 ] := 'Index type must be an ordinal type      ' ;
    cmperror [ 115 ] := 'Base type must be scalar or subrange    ' ;
    cmperror [ 116 ] := 'Error in type of procedure parameter    ' ;
    cmperror [ 117 ] := 'Unsatisfied forward reference           ' ;
    cmperror [ 118 ] := 'Forward reference type identifier       ' ;
    cmperror [ 119 ] := 'Forward declared : repetition par. list ' ;
    cmperror [ 120 ] := 'Function result: scalar,subrange,pointer' ;
    cmperror [ 122 ] := 'Forward declared: repetition result type' ;
    cmperror [ 123 ] := 'Missing result type in function declar. ' ;
    cmperror [ 125 ] := 'Error in type of standard function par. ' ;
    cmperror [ 126 ] := 'Number of parameters disagrees with decl' ;
    cmperror [ 129 ] := 'Incompatible operands                   ' ;
    cmperror [ 130 ] := 'Expression is not of SET type           ' ;
    cmperror [ 131 ] := 'Test on equality allowed only           ' ;
    cmperror [ 132 ] := 'Inclusion not allowed in set comparisons' ;
    cmperror [ 134 ] := 'Illegal type of operands                ' ;
    cmperror [ 135 ] := 'Boolean operands required               ' ;
    cmperror [ 136 ] := 'Set element must be scalar or subrange  ' ;
    cmperror [ 137 ] := 'Set element types not compatible        ' ;
    cmperror [ 138 ] := 'Type must be array                      ' ;
    cmperror [ 139 ] := 'Index type is not compatible with decl. ' ;
    cmperror [ 140 ] := 'Type must be record                     ' ;
    cmperror [ 141 ] := 'Type must be pointer                    ' ;
    cmperror [ 142 ] := 'Illegal parameter substitution          ' ;
    cmperror [ 143 ] := 'Illegal type of loop control variable   ' ;
    cmperror [ 144 ] := 'Illegal type of expression              ' ;
    cmperror [ 145 ] := 'Type conflict                           ' ;
    cmperror [ 147 ] := 'Case label and case expression not comp.' ;
    cmperror [ 148 ] := 'Subrange bounds must be scalar          ' ;
    cmperror [ 149 ] := 'Index type must not be an integer       ' ;
    cmperror [ 150 ] := 'Assignment to standard function illegal ' ;
    cmperror [ 152 ] := 'No such field in this record            ' ;
    cmperror [ 154 ] := 'Actual parameter must be a variable     ' ;
    cmperror [ 155 ] := 'Control variable declared interm. level ' ;
    cmperror [ 156 ] := 'Value already as a label in CASE        ' ;
    cmperror [ 157 ] := 'Too many cases in CASE statement        ' ;
    cmperror [ 160 ] := 'Previous declaration was not forward    ' ;
    cmperror [ 161 ] := 'Again forward declared                  ' ;
    cmperror [ 169 ] := 'SET element not in range 0 .. 63        ' ;
    cmperror [ 170 ] := 'String constant must not exceed one line' ;
    cmperror [ 171 ] := 'Integer constant exceeds range (32767)  ' ;
    cmperror [ 172 ] := 'Too many nested scopes of identifiers   ' ;
    cmperror [ 173 ] := 'Too many nested procedures/functions    ' ;
    cmperror [ 174 ] := 'Index expression out of bounds          ' ;
    cmperror [ 175 ] := 'Internal compiler error : standard funct' ;
    cmperror [ 176 ] := 'Illegal character found                 ' ;
    cmperror [ 177 ] := 'Error in type                           ' ;
    cmperror [ 178 ] := 'Illegal reference to variable           ' ;
    cmperror [ 179 ] := 'Internal error : wrong size variable    ' ;
    cmperror [ 180 ] := 'Maximum number of files exceeded        ' ;

  END ; (* FillErrorMessages *)



PROCEDURE Initialize ;
(********************)

  (* Fills all tables and counters etc with default values *)

  VAR

    i : integer ;

  BEGIN

    errflag   := false ;  (* No errors yet *)
    errtot    := 0 ;
    linecount := 0 ;
    fwptr     := nil   ;
    declpart  := true ;
    prterr    := true ;
    errinx    := 0 ;
    lc        := lcaftermarkstack ;
    ic        := 0 ;
    icn       := 0 ;
    ch        := ' ' ;
    ich       := 32 ;

    listpages := 1 ;
    listlines := maxpage + 1 ;

    mxintio := maxint ;
    nproc := 1 ;

    (* Initialize sets *)

    constbegsys    := [ addop, intconst, stringconst, ident ] ;
    simptypebegsys := [ lparent ] + constbegsys ;
    typebegsys     := [ arrow, packedsy, arraysy, recordsy, setsy ] +
                        simptypebegsys ;
    typedels       := [ arraysy, recordsy, setsy ] ;
    blockbegsys    := [constsy, typesy, varsy, procsy, funcsy, beginsy];
    selectsys      := [ arrow, period, lbrack ] ;
    facbegsys      := [ intconst, stringconst, ident, lparent,
                        realconst, lbrack, notsy ] ;
    statbegsys     := [ beginsy, ifsy, whilesy,
                        repeatsy, casesy, forsy ] ;

    (* Initialize tables *)

    rw [  0 ] := 'if      ' ; rw [  1 ] := 'do      ' ;
    rw [  2 ] := 'of      ' ; rw [  3 ] := 'in      ' ;
    rw [  4 ] := 'or      ' ; rw [  5 ] := 'to      ' ;
    rw [  6 ] := 'end     ' ; rw [  7 ] := 'for     ' ;
    rw [  8 ] := 'var     ' ; rw [  9 ] := 'div     ' ;
    rw [ 10 ] := 'mod     ' ; rw [ 11 ] := 'set     ' ;
    rw [ 12 ] := 'and     ' ; rw [ 13 ] := 'not     ' ;
    rw [ 14 ] := 'then    ' ; rw [ 15 ] := 'else    ' ;
    rw [ 16 ] := 'type    ' ; rw [ 17 ] := 'case    ' ;
    rw [ 18 ] := 'begin   ' ; rw [ 19 ] := 'until   ' ;
    rw [ 20 ] := 'while   ' ; rw [ 21 ] := 'array   ' ;
    rw [ 22 ] := 'const   ' ; rw [ 23 ] := 'repeat  ' ;
    rw [ 24 ] := 'record  ' ; rw [ 25 ] := 'packed  ' ;
    rw [ 26 ] := 'extern  ' ; rw [ 27 ] := 'downto  ' ;
    rw [ 28 ] := 'forward ' ; rw [ 29 ] := 'program ' ;
    rw [ 30 ] := 'function' ; rw [ 31 ] := 'procedur' ;
    rw [ 32 ] := 'otherwis' ;

    frw [ 0 ] := 0   ;
    frw [ 1 ] := 0   ;
    frw [ 2 ] := 6   ;
    frw [ 3 ] := 14  ;
    frw [ 4 ] := 18  ;
    frw [ 5 ] := 23  ;
    frw [ 6 ] := 28  ;
    frw [ 7 ] := 30  ;
    frw [ 8 ] := 33  ;

    (* Initialize symbols *)

    rsy [  0 ] := ifsy      ; rsy [  1 ] := dosy     ;
    rsy [  2 ] := ofsy      ; rsy [  3 ] := relop    ;
    rsy [  4 ] := addop     ; rsy [  5 ] := tosy     ;
    rsy [  6 ] := endsy     ; rsy [  7 ] := forsy    ;
    rsy [  8 ] := varsy     ; rsy [  9 ] := mulop    ;
    rsy [ 10 ] := mulop     ; rsy [ 11 ] := setsy    ;
    rsy [ 12 ] := mulop     ; rsy [ 13 ] := notsy    ;
    rsy [ 14 ] := thensy    ; rsy [ 15 ] := elsesy   ;
    rsy [ 16 ] := typesy    ; rsy [ 17 ] := casesy   ;
    rsy [ 18 ] := beginsy   ; rsy [ 19 ] := untilsy  ;
    rsy [ 20 ] := whilesy   ; rsy [ 21 ] := arraysy  ;
    rsy [ 22 ] := constsy   ; rsy [ 23 ] := repeatsy ;
    rsy [ 24 ] := recordsy  ; rsy [ 25 ] := packedsy ;
    rsy [ 26 ] := externsy  ; rsy [ 27 ] := downtosy ;
    rsy [ 28 ] := forwardsy ; rsy [ 29 ] := progsy   ;
    rsy [ 30 ] := funcsy    ; rsy [ 31 ] := procsy   ;
    rsy [ 32 ] := elsesy    ;

    for i :=0 to 127 DO
      BEGIN
        ssy [ i ]  := othersy ;
        sop [ i ]  := noop
      END ;

    ssy [ ord('+') ] := addop     ;
    ssy [ ord('-') ] := addop     ;
    ssy [ ord('*') ] := mulop     ;
    ssy [ ord('/') ] := mulop     ;
    ssy [ ord('(') ] := lparent   ;
    ssy [ ord(')') ] := rparent   ;
    ssy [ ord('$') ] := othersy   ;
    ssy [ ord('=') ] := relop     ;
    ssy [ ord(' ') ] := othersy   ;
    ssy [ ord(',') ] := comma     ;
    ssy [ ord('.') ] := period    ;
    ssy [ ord('"') ] := othersy   ;
    ssy [ ord('[') ] := lbrack    ;
    ssy [ ord(']') ] := rbrack    ;
    ssy [ ord(':') ] := colon     ;
    ssy [ ord('^') ] := arrow     ;
    ssy [ ord('<') ] := relop     ;
    ssy [ ord('>') ] := relop     ;
    ssy [ ord(';') ] := semicolon ;

    (* Initialize operators *)

    for i := 0 to 32 DO
      rop [ i ] := noop ;
    rop [  3 ] := inop  ;
    rop [  4 ] := orop  ;
    rop [  9 ] := idiv  ;
    rop [ 10 ] := imod  ;
    rop [ 12 ] := andop ;

    sop [ ord('+') ] := plus  ;
    sop [ ord('-') ] := minus ;
    sop [ ord('*') ] := mul   ;
    sop [ ord('=') ] := eqop  ;
    sop [ ord('/') ] := idiv  ;
    sop [ ord('<') ] := ltop  ;
    sop [ ord('>') ] := gtop  ;

    (* Enter standard names and standard types *)

    level      := 0   ;
    top        := 0   ;
    for i := 0 to displimit DO
      display[i] := nil ;

    Enterstdtypes  ;
    Stdnames ;
    Enterstnames ;
    Enterundecl ;

    top := 1 ;
    level := 1 ;
    savetop := 1 ;

    (* filenumber start with 4 : input = 1, output = 2, keyboard := 3 *)

    filenumber := 4 ;

    (* Fill error-message table *)
    FillErrorMessages ;

  END ; (* Initialize *)


  PROCEDURE OpenFiles ;

      VAR

         i : integer ;
         filename,
         filenamepart : string ;

       BEGIN (* Openfiles *)

         if paramcount = 0
           then
             begin
               write('Pascal-M object file to load> ') ;
               readln(filename)
             end
           else
             filename := paramstr(1) ;
         if FileExists(filename)
           then
             begin
               assign (infile, filename);
               reset(infile) ;
             end
            else
              begin
                writeln('No such file') ;
                halt(1) ;
              end;

         { extract filename part }
         filenamepart := '' ;
         i := 1 ;
         while filename[i] <> '.' do
           begin
             filenamepart := filenamepart + filename[i] ;
             i := i + 1 ;
           end ;
       filename := filenamepart + '.obp' ;
       assign (outfile, filename) ;
       rewrite(outfile) ;
       filename := filenamepart + '.lst' ;
       assign( listfile, filename) ;
       rewrite(listfile) ;

   END ; (* OpenFiles *)

PrOCEDURE CloseFiles ;

BEGIN

  close(infile) ;
  close(outfile) ;
  close(listfile) ;

END ; (* CloseFiles *)


  BEGIN (* main Mpascal *)

    Initialize ;
    writeln ( 'Pascal-M compiler V1.4 Freepascal version August 2020') ;
    writeln ;

    Openfiles ;

   (* Compile program-heading *)

    BeginLine ;
    InSymbol ;
    new (progptr ) ;
    progptr^.idtype := nil ;
    progptr^.next := nil ;
    progptr^.klass := proc ;
    progptr^.pfdeckind := declared ;
    progptr^.pfname := 0 ;
    progptr^.name := '        ' ;

    IF sy = progsy
      THEN
        BEGIN
          (* program name *)
          InSymbol ;
          IF sy <> ident
            THEN
              Error(2)
            ELSE
              progptr^.name := id ;

          (* Skip input/output after program name *)

          WHILE sy <> semicolon DO
            InSymbol ;

          (* Skip semicolon *)

          InSymbol ;

        END ;

    (* jump into block untill end *)
    Block (blockbegsys + statbegsys, period , progptr ) ;

    EndLine ;
    WriteOut ;
    IF errflag
      THEN
        BEGIN
          writeln ;
          EndLineList ;
          writeln (listfile) ;
          writeln ;
          writeln (' Pascal-M Compilation : ', errtot:4,' errors ' ) ;
          EndLineList ;
          write ( listfile, ' Pascal-M Compilation : ') ;
          writeln ( listfile, errtot:4,' errors ' ) ;
          writeln(outfile,'P1010000');
        END
      ELSE
        BEGIN
          writeln ;
          EndLineList ;
          writeln (listfile) ;
          writeln ('Pascal-M Compilation successful' ) ;
          EndLineList ;
          writeln (listfile, ' Pascal-M Compilation successful' ) ;
          writeln( outfile, 'P9' ) ;
        END ;
   CloseFiles ;
  END.

